from .basic_alg import arctan2, one_degree, sqrt, cos, sin, matmul, arccos, rref, nan


def get_distance_by_tow_point(x1, y1, x2, y2):
    return sqrt((x1 - x2) ** 2 + (y1 - y2) ** 2)


def radian2angle(r):
    return r / one_degree


def angle(v1x, v1y, v2x, v2y):
    return arctan2(v1y - v2y, v1x - v2x)


def polar2descartes(r, theta):
    return r * cos(theta), r * sin(theta)


def descartes2polar(x, y):
    return sqrt(x ** 2 + y ** 2), arctan2(y, x)


def polar(geo):
    old_set_data = geo.set_data

    def new_set_data(r, t):
        x, y = polar2descartes(t, r)
        old_set_data(x, y)

    geo.set_data = new_set_data
    geo.update_self()


def get_angle_by_three_point_corr(x1, y1, x2, y2, x3, y3):
    v1 = [x1 - x2, y1 - y2]
    v2 = [x3 - x2, y3 - y2]
    v1_v2 = v1[0] * v2[0] + v1[1] * v2[1]
    v1_len = get_distance_by_tow_point(x1, y1, x2, y2)
    v2_len = get_distance_by_tow_point(x3, y3, x2, y2)
    cos_a = v1_v2 / (v1_len * v2_len)
    a = arccos(cos_a)
    s = v1[0] * v2[1] - v1[1] * v2[0]

    if s > 0:
        return a
    return -a


def get_intersect_point_corr_by_four_point(x1, y1, x2, y2, x3, y3, x4, y4):
    """
    line 1 : (x1, y1), (x2, y2)
    line 2 : (x3, y3), (x4, y4)
    """
    a11 = y2 - y1
    a12 = x1 - x2
    a21 = y4 - y3
    a22 = x3 - x4
    b1 = a11 * x1 + a12 * y1
    b2 = a21 * x3 + a22 * y3
    mat = [
        [a11, a12, b1],
        [a21, a22, b2]
    ]
    mat_simple = rref(mat)
    if mat_simple[1][1] != 0:
        return [mat_simple[0][2], mat_simple[1][2]]
    if mat_simple[1][2] != 0:
        return [nan, nan]
    return [x1, y1]


def get_intersect_point_corr_by_four_point_segment_and_segment(x1, y1, x2, y2, x3, y3, x4, y4):
    ix, iy = get_intersect_point_corr_by_four_point(x1, y1, x2, y2, x3, y3, x4, y4)
    if ix == nan:
        return [nan, nan]

    min_x_12 = min(x1, x2)
    min_y_12 = min(y1, y2)
    max_x_12 = max(x1, x2)
    max_y_12 = max(y1, y2)
    min_x_34 = min(x3, x4)
    min_y_34 = min(y3, y4)
    max_x_34 = max(x3, x4)
    max_y_34 = max(y3, y4)

    if not \
            min_x_12 <= ix <= max_x_12 or not \
            min_x_34 <= ix <= max_x_34 or not \
            min_y_12 <= iy <= max_y_12 or not \
            min_y_34 <= iy <= max_y_34:
        return [nan, nan]

    return [ix, iy]


def get_intersect_point_corr_by_four_point_line_and_segment(x1, y1, x2, y2, x3, y3, x4, y4):
    ix, iy = get_intersect_point_corr_by_four_point(x1, y1, x2, y2, x3, y3, x4, y4)
    if ix == nan:
        return [nan, nan]

    min_x_34 = min(x3, x4)
    min_y_34 = min(y3, y4)
    max_x_34 = max(x3, x4)
    max_y_34 = max(y3, y4)

    if not \
            min_x_34 <= ix <= max_x_34 or not \
            min_y_34 <= iy <= max_y_34:
        return [nan, nan]

    return [ix, iy]


def get_intersect_point_corr_by_circle_and_circle(x, y, R, a, b, S):
    d = sqrt((abs(a - x)) ** 2 + (abs(b - y)) ** 2)
    if d > (R + S) or d < (abs(R - S)):
        return [nan, nan]
    elif d == 0:
        return [nan, nan]
    else:
        A = (R ** 2 - S ** 2 + d ** 2) / (2 * d)
        h = sqrt(R ** 2 - A ** 2)
        x2 = x + A * (a - x) / d
        y2 = y + A * (b - y) / d
        x3 = round(x2 - h * (b - y) / d, 2)
        y3 = round(y2 + h * (a - x) / d, 2)
        x4 = round(x2 + h * (b - y) / d, 2)
        y4 = round(y2 - h * (a - x) / d, 2)
        c1 = [x3, y3]
        c2 = [x4, y4]
        return c1, c2


def get_edge_two_point_by_line(x1, y1, x2, y2, x_min, x_max, y_min, y_max):
    if x1 == x2:
        return [[x1, y_min], [x1, y_max]]

    k = (y1 - y2) / (x1 - x2)
    y1_k_x1 = y1 - k * x1
    y_min = k * x_min + y1_k_x1
    y_max = k * x_max + y1_k_x1

    return [[x_min, y_min], [x_max, y_max]]


def get_abc_by_line(x1, y1, x2, y2):
    if y1 == y2:
        return [0, 1, -y1]
    if x1 == x2:
        return [1, 0, -x1]
    x1_x2 = x1 - x2
    y1_y2 = y1 - y2
    return [y1 - y2, x1 - x2, x1_x2 * y1 - y1_y2 * x1]


def get_middle_point_by_two_point(x1, y1, x2, y2):
    return (x1 + x2) / 2, (y1 + y2) / 2


def get_rotate_point_by_two_point(x1, y1, x2, y2, t):
    x_2_1 = x2 - x1
    y_2_1 = y2 - y1
    return [x1 + x_2_1 * cos(t) - y_2_1 * sin(t),
            y1 + x_2_1 * sin(t) + y_2_1 * cos(t)]


def get_vertical_point_by_point_line(px, py, x1, y1, x2, y2):
    x2_x1 = x2 - x1
    y2_y1 = y2 - y1
    k = ((px - x1) * x2_x1 + (py - y1) * y2_y1) / (x2_x1 ** 2 + y2_y1 ** 2)
    return [x1 + k * x2_x1, y1 + k * y2_y1]


def get_symmetric_point_by_point_line(px, py, x1, y1, x2, y2):
    vp = get_vertical_point_by_point_line(px, py, x1, y1, x2, y2)
    return [2 * vp[0] - px, 2 * vp[1] - py]


def get_symmetric_point_by_point(x1, y1, x2, y2):
    return [2 * x2 - x1, 2 * y2 - y1]


def get_point_by_matrix(mat, x1, y1):
    return matmul(mat, [x1, y1])


def get_transform_vector_to_y_up(x1, y1, x2, y2):
    """
    Move (x1, y1) to positive y-axis become (x3, y3), move (x2, y2) to negative y-axis become (x4, y4),
    and absolute y-coordinate is equal.

    return : [t1, t2]
    t1: [(x3, y3), (x4, y4)] transform to [(x1, y1), (x2, y2)]
    """
    x0, y0 = (x1 + x2) / 2, (y1 + y2) / 2
    delta_x = x1 - x2
    delta_y = y1 - y2
    if delta_x == 0:
        if delta_y == 0:
            return 2 * [lambda x, y: [x, y]]
        if delta_y < 0:
            return [lambda x, y: [-x + x0, -y + y0], lambda x, y: [x0 - x, y0 - y]]
        return [lambda x, y: [x + x0, y + y0], lambda x, y: [x - x0, y - y0]]
    if delta_y == 0:
        if delta_x < 0:
            return [lambda x, y: [-y + x0, x + y0], lambda x, y: [(y - y0), -(x - x0)]]
        return [lambda x, y: [y + x0, -x + y0], lambda x, y: [-(y - y0), (x - x0)]]
    tan_alpha = delta_x / delta_y
    sqrt_1_add_tan2_alpha = sqrt(1 + tan_alpha ** 2)
    if delta_y < 0:
        sqrt_1_add_tan2_alpha = -sqrt_1_add_tan2_alpha
    sin_alpha = tan_alpha / sqrt_1_add_tan2_alpha
    cos_alpha = 1 / sqrt_1_add_tan2_alpha
    return [lambda x, y: [cos_alpha * x + sin_alpha * y + x0, -sin_alpha * x + cos_alpha * y + y0],
            lambda x, y: [cos_alpha * (x - x0) - sin_alpha * (y - y0), sin_alpha * (x - x0) + cos_alpha * (y - y0)]]



from matplotlib.animation import FuncAnimation
from .king_figure import KingFigure
from ..alg.basic_alg import linspace


class KingAnimation:
    """
    Animation.
    Parameter:
    ----------
    artists : must be KingGeometry and Artist's instance list.
    essential_data_generators : A list of function which generates essential_data for each artists,
                                accept a parameter represent for frame.
    frames : Iterable or callable.
    interval : Time interval.
    See FuncAnimation.
    """

    def __init__(self, artists, essential_data_generators, frames=linspace(0, 10, 101), interval=50):
        KingFigure().update()
        self.artists = artists
        self.essential_data_generators = essential_data_generators
        self.frames = frames
        self.interval = interval
        self.count = len(artists)

        for a in artists:
            a.mouse_translation = False
        self.ani = FuncAnimation(KingFigure().fig, self.update, frames=frames, interval=interval, blit=True)

    def update(self, frame):
        result = []

        for i in range(self.count):
            result.extend(self.artists[i].update_affected(new_essential_data=self.essential_data_generators[i](frame),
                                                          repaint=False))
        KingFigure().update()
        return result



from .basic_alg import rref
from .plane_geometry_alg import get_distance_by_tow_point


def get_orthocenter_by_three_point(x1, y1, x2, y2, x3, y3):
    a11 = x2 - x1
    a12 = y2 - y1
    a21 = x2 - x3
    a22 = y2 - y3
    b1 = a11 * x3 + a12 * y3
    b2 = a21 * x1 + a22 * y1

    mat = [
        [a11, a12, b1],
        [a21, a22, b2]
    ]

    rref_mat = rref(mat)

    return [rref_mat[0][2], rref_mat[1][2]]


def get_meidan_by_three_point(x1, y1, x2, y2, x3, y3):
    x4 = (x1 + x2) / 2
    y4 = (y1 + y2) / 2
    return [x3 / 3 + 2 * x4 / 3, y3 / 3 + 2 * y4 / 3]


def get_incenter_by_three_point(x1, y1, x2, y2, x3, y3):
    a = get_distance_by_tow_point(x2, y2, x3, y3)
    b = get_distance_by_tow_point(x1, y1, x3, y3)
    c = get_distance_by_tow_point(x1, y1, x2, y2)
    abc = a + b + c

    return [(a * x1 + b * x2 + c * x3) / abc, (a * y1 + b * y2 + c * y3) / abc]


def get_outer_by_three_point(x1, y1, x2, y2, x3, y3):
    x4 = (x1 + x2) / 2
    y4 = (y1 + y2) / 2
    x5 = (x1 + x3) / 2
    y5 = (y1 + y3) / 2
    a11 = x1 - x2
    a12 = y1 - y2
    a21 = x1 - x3
    a22 = y1 - y3
    b1 = a11 * x4 + a12 * y4
    b2 = a21 * x5 + a22 * y5
    mat = [
        [a11, a12, b1],
        [a21, a22, b2]
    ]
    rref_mat = rref(mat)

    return [rref_mat[0][2], rref_mat[1][2]]



import matplotlib.pyplot as plt
from .king_decorater import singleton


@singleton
class KingFigure:
    def __init__(self):
        # plt.rcParams['toolbar'] = 'none'
        self.fig = plt.figure()
        self.ax = self.fig.add_axes([0, 0, 1, 1])
        self.canvas = self.fig.canvas
        self.artists = self.ax.artists
        self.background = None
        self.scroll_rate = 0.1
        self.epsilon = 0.2

        self.old_x = None
        self.old_y = None
        self.selected_geometry_list = []
        self.is_selected_geometry = None

        # The geometry has a infinite bound. such as line, hyperbola, etc.
        # it should be repaint when the axis's limit changed.
        self.inf_geometry = []

        # Whether delay update when create a geometry.
        self.init_delay_update = False

        # self.canvas.manager.window.showMaximized()
        self.ax.spines['top'].set_color('none')
        self.ax.spines['right'].set_color('none')
        self.ax.xaxis.set_ticks_position('bottom')
        self.ax.spines['bottom'].set_position(('data', 0))
        self.ax.yaxis.set_ticks_position('left')
        self.ax.spines['left'].set_position(('data', 0))
        x_lim = y_lim = 20
        self.set_x_lim(-x_lim, x_lim)
        self.set_y_lim(-y_lim, y_lim)
        self.ax.set_aspect(1)

        self.ax.grid()

        self.canvas.mpl_connect("scroll_event", self.on_scroll)
        self.canvas.mpl_connect("draw_event", self.on_draw)
        self.canvas.mpl_connect('pick_event', self.on_pick)
        self.canvas.mpl_connect('button_press_event', self.on_button_press)
        self.canvas.mpl_connect('button_release_event', self.on_button_release)
        self.canvas.mpl_connect('motion_notify_event', self.on_mouse_move)

        self.ax.callbacks.connect('xlim_changed', self.on_lim_changed)

    def show(self):
        self.fig.show()

    def on_lim_changed(self, event):
        for a in self.inf_geometry:
            a.update_self()

    def set_x_lim(self, x_min, x_max):
        self.ax.set_xlim(left=x_min, right=x_max)

    def set_y_lim(self, y_min, y_max):
        self.ax.set_ylim(bottom=y_min, top=y_max)

    def get_x_lim(self):
        return self.ax.get_xlim()

    def get_y_lim(self):
        return self.ax.get_ylim()

    def on_scroll(self, event):
        if event.inaxes is None:
            return
        x_min, x_max = self.get_x_lim()
        y_min, y_max = self.get_y_lim()
        x_mouse, y_mouse = event.xdata, event.ydata
        rate = self.scroll_rate if event.button == 'up' else -self.scroll_rate
        delta_l = (x_mouse - x_min) * rate
        delta_r = (x_max - x_mouse) * rate
        delta_u = (y_max - y_mouse) * rate
        delta_d = (y_mouse - y_min) * rate
        x_min_new, x_max_new = x_min + delta_l, x_max - delta_r
        y_min_new, y_max_new = y_min + delta_d, y_max - delta_u
        self.epsilon *= (1 - rate)
        self.set_x_lim(x_min_new, x_max_new)
        self.set_y_lim(y_min_new, y_max_new)
        self.canvas.draw_idle()

    def add_artist(self, artist):
        self.ax.add_artist(artist)

    def do_draw(self):
        for a in self.artists:
            self.ax.draw_artist(a)

    def on_draw(self, event):
        self.background = self.canvas.copy_from_bbox(self.ax.bbox)
        self.do_draw()

    def update(self):
        if self.background is None:
            self.canvas.draw()
        else:
            self.canvas.restore_region(self.background)
            self.do_draw()
            self.canvas.blit(self.ax.bbox)
        self.canvas.flush_events()

    def on_pick(self, event):
        self.selected_geometry_list.append(event.artist)
        for a in self.selected_geometry_list:
            if self.is_selected_geometry is None:
                self.is_selected_geometry = a
                continue
            if a.mouse_select_priority < self.is_selected_geometry.mouse_select_priority:
                self.is_selected_geometry = a

    def on_button_press(self, event):
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        self.old_x = event.xdata
        self.old_y = event.ydata

    def on_mouse_move(self, event):
        if event.inaxes is None:
            return
        if event.button != 1:
            return
        if self.is_selected_geometry is None:
            return
        if not self.is_selected_geometry.mouse_translation:
            return
        delta_x = event.xdata - self.old_x
        delta_y = event.ydata - self.old_y
        self.old_x = event.xdata
        self.old_y = event.ydata
        self.is_selected_geometry.translation((delta_x, delta_y), (event.xdata, event.ydata))
        self.update()

    def on_button_release(self, event):
        self.selected_geometry_list = []
        self.is_selected_geometry = None

    def clear_searched(self):
        for a in self.artists:
            a.searched = False



from ..core.king_figure import KingFigure
from .king_curve import KingCurve
from ..alg.basic_alg import linspace


class KingFunction(KingCurve):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        """
        {'function': f, 'x_min': min, 'x_max': max, 'count': c}
        """
        super(KingFunction, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                           infinite=True)

    def generate_basic_data(self, depends=(), essential_data=None):
        return essential_data

    def generate_xy_data_from_basic_data(self):
        x_min = self.basic_data['x_min']
        x_max = self.basic_data['x_max']
        count = self.basic_data['count']
        x_min_figure, x_max_figure = KingFigure().get_x_lim()
        if x_min is None:
            x_min = x_min_figure
        if x_max is None:
            x_max = x_max_figure
        if count is None:
            count = 100
        x = linspace(x_min, x_max, count)
        return [x, self.basic_data['function'](x)]



from ..core.king_decorater import geometry_generate
from .king_plane_geometry_graph import KingPoint
from ..alg.triangle_center_alg import *


@geometry_generate(cls=KingPoint)
def orthocenter(depends=(), essential_data=None):
    p1 = depends[0]
    p2 = depends[1]
    p3 = depends[2]

    return get_orthocenter_by_three_point(*p1.basic_data, *p2.basic_data, *p3.basic_data)


@geometry_generate(cls=KingPoint)
def median(depends=(), essential_data=None):
    p1 = depends[0]
    p2 = depends[1]
    p3 = depends[2]

    return get_meidan_by_three_point(*p1.basic_data, *p2.basic_data, *p3.basic_data)


@geometry_generate(cls=KingPoint)
def incenter(depends=(), essential_data=None):
    p1 = depends[0]
    p2 = depends[1]
    p3 = depends[2]

    return get_incenter_by_three_point(*p1.basic_data, *p2.basic_data, *p3.basic_data)


@geometry_generate(cls=KingPoint)
def outer(depends=(), essential_data=None):
    p1 = depends[0]
    p2 = depends[1]
    p3 = depends[2]

    return get_outer_by_three_point(*p1.basic_data, *p2.basic_data, *p3.basic_data)



from .basic_alg import sqrt, get_main_element_index_list, real_root_tow_power_equation, det, \
    is_same_negative_or_positive, nan, rref, eig_n, pi_half, is_zero, less_than_zero, greater_than_zero
from .circle_alg import get_circle_equation_coefficient_by_center_radius
from .plane_geometry_alg import get_rotate_point_by_two_point
from .parabola_alg import get_parabola_equation_coefficient_by_focus_line

CONIC_TYPE_POINT = 0
CONIC_TYPE_LINE = 1
CONIC_TYPE_PARALLEL_LINE = 2
CONIC_TYPE_INTERSECT_LINE = 3
CONIC_TYPE_ELLIPSE = 4
CONIC_TYPE_CIRCLE = 5
CONIC_TYPE_HYPERBOLA = 6
CONIC_TYPE_PARABOLA = 7
CONIC_TYPE_CAN_NOT_PLOT = 8


def elimination_of_cross_phase_by_rotating_shaft(a11, a12, a22, a1, a2, a0):
    """
    a11*x^2 + 2*a12*xy + a22*y^2 + 2a1 * x + 2a2 * y + a0 = 0
    (x, y)' = A * (x2, y2)'
    b11*x2^2 + b22*y2^2 + 2b1 * x2 + 2b2 * y2 + b0 = 0

    return [A, [b11, 0, b22, b1, b2, b0]]
    """
    if a12 == 0:
        mat = [
            [1, 0],
            [0, 1]
        ]
        return mat, [a11, a12, a22, a1, a2, a0]
    delta = (a11 - a22) ** 2 + 4 * a12 ** 2
    sqrt_delta = sqrt(delta)
    sqrt_delta_2 = sqrt_delta / 2
    a11_a22_2 = (a11 + a22) / 2

    eigenvalue_1 = a11_a22_2 + sqrt_delta_2
    eigenvector_1_x = 1
    eigenvector_1_y = (eigenvalue_1 - a11) / a12
    len_1 = sqrt(1 + eigenvector_1_y ** 2)
    eigenvector_1_x /= len_1
    eigenvector_1_y /= len_1

    eigenvalue_2 = a11_a22_2 - sqrt_delta_2
    eigenvector_2_x = 1
    eigenvector_2_y = (eigenvalue_2 - a11) / a12
    len_2 = sqrt(1 + eigenvector_2_y ** 2)
    eigenvector_2_x /= len_2
    eigenvector_2_y /= len_2

    mat = [
        [eigenvector_1_x, eigenvector_2_x],
        [eigenvector_1_y, eigenvector_2_y]
    ]
    b1 = a1 * eigenvector_1_x + a2 * eigenvector_1_y
    b2 = a1 * eigenvector_2_x + a2 * eigenvector_2_y

    return mat, [eigenvalue_1, 0, eigenvalue_2, b1, b2, a0]


def shift_axis(a11, a12, a22, a1, a2, a0):
    """
    a11*x^2 + a22*y^2 + 2a1 * x + 2a2 * y + a0 = 0
    (x, y)' = (x2, y2)' + B
    b11*x2^2 + b22*y2^2 + 2b1 * x2 + 2b2 * y2 + b0 = 0

    return [B, [b11, 0, b22, b1, b2, b0]]
    """
    if a11 != 0:
        if a22 != 0:
            return [[a1 / a11, a2 / a22], [a11, 0, a22, 0, 0, a0 - a1 ** 2 / a11 - a2 ** 2 / a22]]
        if a2 == 0:
            return [[a1 / a11, 0], [a11, 0, 0, 0, 0, a0 - a1 ** 2 / a11]]
        return [[a1 / a11, (a0 - a1 ** 2 / a11) / (2 * a2)], [a11, 0, 0, 0, a2, 0]]
    if a22 != 0:
        if a1 == 0:
            return [[0, a2 / a22], [0, 0, a22, 0, 0, a0 - a2 ** 2 / a22]]
        return [[(a0 - a2 ** 2 / a22) / (2 * a1), a2 / a22], [0, 0, a22, a1, 0, 0]]
    return [[0, 0], [0, 0, 0, a1, a2, a0]]


def get_simplest_conic_equation_and_transform_by_conic_equation(a11, a12, a22, a1, a2, a0):
    """
    a11*x^2 + 2*a12*xy + a22*y^2 + 2a1 * x + 2a2 * y + a0 = 0

    X1 = (x1, y1)'
    X2 = (x2, y2)'
    X3 = (x3, y3)'

    X1 = A * X2
    X3 = B + X2

    return [t1, t2, [c11, c12, c22, c1, c2, c0]]
    t1 : X3 -> X1 : X1 = A * (X3 - B)
    t2 : X1 -> X3 : X3 = A^(-1) * X1 + B
    """
    rotate_mat, rotate_coefficient = elimination_of_cross_phase_by_rotating_shaft(a11, a12, a22, a1, a2, a0)
    translation_mat, translation_coefficient = shift_axis(*rotate_coefficient)

    def t1(x, y):
        if type(x) == list:
            return list(zip(*[t1(x[i], y[i]) for i in range(len(x))]))
        x_translation_mat_0 = x - translation_mat[0]
        y_translation_mat_1 = y - translation_mat[1]
        return [rotate_mat[0][0] * x_translation_mat_0 + rotate_mat[0][1] * y_translation_mat_1,
                rotate_mat[1][0] * x_translation_mat_0 + rotate_mat[1][1] * y_translation_mat_1]

    def t2(x, y):
        if type(x) == list:
            return list(zip(*[t2(x[i], y[i]) for i in range(len(x))]))
        return [rotate_mat[0][0] * x + rotate_mat[1][0] * y + translation_mat[0],
                rotate_mat[0][1] * x + rotate_mat[1][1] * y + translation_mat[1]]

    return [t1, t2, translation_coefficient]


def get_conic_type_and_basic_data_by_simplest_equation(a11, a12, a22, a1, a2, a0):
    """
    get conic type and basic data by simplest equation.

    return [type, basic_data]
    """
    if less_than_zero(a11):
        a11 = -a11
        a22 = -a22
        a0 = -a0
    if greater_than_zero(a11):
        if greater_than_zero(a22):
            if less_than_zero(a0):
                if greater_than_zero(a22-a11):
                    cc = (1 / a22 - 1 / a11) * a0
                    c = sqrt(cc)
                    return [CONIC_TYPE_ELLIPSE, [[c, 0], [-c, 0], 2 * sqrt(-a0 / a11)]]
                elif is_zero(a22-a11):
                    return [CONIC_TYPE_CIRCLE, sqrt(-a0 / a11)]
                else:
                    cc = (1 / a11 - 1 / a22) * a0
                    c = sqrt(cc)
                    return [CONIC_TYPE_ELLIPSE, [[0, c], [0, -c], 2 * sqrt(-a0 / a22)]]
            elif greater_than_zero(a0):
                return [CONIC_TYPE_CAN_NOT_PLOT]
            else:
                # (0, 0)
                return [CONIC_TYPE_POINT]
        elif less_than_zero(a22):
            if is_zero(a0):
                k = sqrt(-a11 / a22)
                return [CONIC_TYPE_INTERSECT_LINE, [[[0, 0], [1, k]], [[0, 0], [1, -k]]]]
            elif greater_than_zero(a0):
                cc = (1 / a11 - 1 / a22) * a0
                c = sqrt(cc)
                return [CONIC_TYPE_HYPERBOLA, [[0, c], [0, -c], 2 * sqrt(-a0 / a22)]]
            else:
                cc = (1 / a22 - 1 / a11) * a0
                c = sqrt(cc)
                return [CONIC_TYPE_HYPERBOLA, [[c, 0], [-c, 0], 2 * sqrt(-a0 / a11)]]
        else:
            if is_zero(a2):
                if greater_than_zero(a0):
                    return [CONIC_TYPE_CAN_NOT_PLOT]
                elif less_than_zero(a0):
                    x = sqrt(-a0 / a11)
                    return [CONIC_TYPE_PARALLEL_LINE, [[[x, 0], [x, 1]], [[-x, 0], [-x, 1]]]]
                else:
                    return [CONIC_TYPE_LINE, [[0, 0], [0, 1]]]
            else:
                return [CONIC_TYPE_PARABOLA, [[0, -(a0 / a2 + a2 / a11) / 2], [0, (a2 / a11 - a0 / a2) / 2]]]
    else:
        if is_zero(a22):
            if is_zero(a1):
                return [CONIC_TYPE_CAN_NOT_PLOT]
            x = -a0 / (2 * a1)
            return [CONIC_TYPE_LINE, [[x, 0], [x, 1]]]
        if less_than_zero(a22):
            a22 = -a22
            a1 = -a1
            a0 = -a0
        if greater_than_zero(a22):
            if is_zero(a1):
                if is_zero(a0):
                    return [CONIC_TYPE_LINE, [[0, 0], [1, 0]]]
                elif greater_than_zero(a0):
                    return [CONIC_TYPE_CAN_NOT_PLOT]
                else:
                    y = sqrt(-a0 / a22)
                    return [CONIC_TYPE_PARALLEL_LINE, [[[0, y], [1, y]], [[0, -y], [1, -y]]]]
            else:
                return [CONIC_TYPE_PARABOLA, [[-(a1 / a22 + a0 / a1) / 2, 0], [(a1 / a22 - a0 / a1) / 2, 0]]]


def get_conic_mat_by_five_point(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5):
    """
    a11 * x^2 + 2a12 * xy + a22 * y^2 + 2a1 * x + 2a2 * y + a0 = 0
    """

    def row(x, y):
        return [x ** 2, 2 * x * y, y ** 2, 2 * x, 2 * y, 1]

    return [row(*point) for point in zip([x1, x2, x3, x4, x5], [y1, y2, y3, y4, y5])]


def get_conic_coefficient(row_simplest_matrix):
    """
    Parameter
    =========
    row simplest matrix represent for 5 point conic's equation.

    return: A list which elements is: [a11, a12, a22, a1, a2, a0]
    """
    index_list = get_main_element_index_list(row_simplest_matrix)
    rank = len(index_list)
    if rank == 5:
        i = 0
        flag = True
        while i < rank - 1:
            if index_list[i + 1] - index_list[i] > 1:
                i += 1
                flag = False
                break
            i += 1
        if flag:
            i = rank
        result = [row_simplest_matrix[j][i] for j in range(i)]
        result.append(-1)
        result.extend([0] * (rank - i))
        return result
    # TODO : case rank < 5


def get_central_conic_equation_coefficient_by_focus_d(x1, y1, x2, y2, d):
    """
    central conic include ellipse, circle, hyperbola.
    especially ellipse and hyperbola.
    when ellipse, assign two focus and add.
    when hyperbola, assign two focus and difference.
    """
    dd = d * d
    dddd = dd * dd
    x1x1 = x1 * x1
    x2x2 = x2 * x2
    y1y1 = y1 * y1
    y2y2 = y2 * y2
    x1x1x1x1 = x1x1 * x1x1
    x2x2x2x2 = x2x2 * x2x2
    y1y1y1y1 = y1y1 * y1y1
    y2y2y2y2 = y2y2 * y2y2
    x1_x2 = x1 - x2
    y1_y2 = y1 - y2
    x1_add_x2 = x1 + x2
    y1_add_y2 = y1 + y2
    x1_x2_2 = x1_x2 * x1_x2
    y1_y2_2 = y1_y2 * y1_y2
    a11 = 4 * (x1_x2_2 - dd)
    a22 = 4 * (y1_y2_2 - dd)
    a12 = 4 * x1_x2 * y1_y2
    a1 = 2 * (dd * x1_add_x2 - x1_add_x2 * x1_x2_2 - x1_x2 * y1_add_y2 * y1_y2)
    a2 = 2 * (dd * y1_add_y2 - y1_add_y2 * y1_y2_2 - y1_y2 * x1_add_x2 * x1_x2)
    a0 = x1x1x1x1 - 2 * x1x1 * x2x2 + 2 * x1x1 * y1y1 - 2 * x1x1 * y2y2 + x2x2x2x2 - 2 * x2x2 * y1y1 + 2 * x2x2 * y2y2 + y1y1y1y1 - 2 * y1y1 * y2y2 + y2y2y2y2 + dddd - 2 * dd * x1x1 - 2 * dd * x2x2 - 2 * dd * y1y1 - 2 * dd * y2y2

    return [a11, a12, a22, a1, a2, a0]


def get_conic_value_by_point(x0, y0, a11, a12, a22, a1, a2, a0):
    return a11 * x0 ** 2 + 2 * a12 * x0 * y0 + a22 * y0 ** 2 + 2 * a1 * x0 + 2 * a2 * y0 + a0


def get_conic_I1(a11, a12, a22, a1, a2, a0):
    return a11 + a22


def get_conic_I2(a11, a12, a22, a1, a2, a0):
    return a11 * a22 - a12 ** 2


def get_conic_I3(a11, a12, a22, a1, a2, a0):
    mat = [
        [a11, a12, a1],
        [a12, a22, a2],
        [a1, a2, a0]
    ]
    return det(mat)


def get_conic_K1(a11, a12, a22, a1, a2, a0):
    return a0 * (a11 + a22) - a1 ** 2 - a2 ** 2


def get_conic_type_simply_equation(a11, a12, a22, a1, a2, a0):
    I1 = get_conic_I1(a11, a12, a22, a1, a2, a0)
    I2 = get_conic_I2(a11, a12, a22, a1, a2, a0)
    I3 = get_conic_I3(a11, a12, a22, a1, a2, a0)

    if is_zero(I2):
        if is_zero(I3):
            K1 = get_conic_K1(a11, a12, a22, a1, a2, a0)
            if is_zero(K1):
                return [CONIC_TYPE_LINE, [0, 0, 1, 0, 0, 0]]
            if K1 < 0:
                return [CONIC_TYPE_PARALLEL_LINE, [0, 0, 1, 0, 0, K1 / (I1 ** 2)]]
            if K1 > 0:
                return [CONIC_TYPE_CAN_NOT_PLOT]
        return [CONIC_TYPE_PARABOLA, [0, 0, I1, 2 * (-I3 / I1) ** (1 / 2), 0, 0]]
    if I2 > 0:
        same_symbol = is_same_negative_or_positive(I1, I3)
        if same_symbol is not None:
            return [CONIC_TYPE_CAN_NOT_PLOT]
        if is_zero(I3):
            return [CONIC_TYPE_POINT]
        roots = real_root_tow_power_equation(1, -I1, I2)
        if len(roots) == 1:
            return [CONIC_TYPE_CIRCLE, [1, 0, 1, 0, 0, I3 / I2 / roots[0]]]
        return [CONIC_TYPE_ELLIPSE, [roots[0], 0, roots[1], 0, 0, I3 / I2]]
    if I2 < 0:
        roots = real_root_tow_power_equation(1, -I1, I2)
        root_1 = roots[0]
        root_2 = root_1 if len(roots) == 1 else roots[1]
        corr = [root_1, 0, root_2, 0, 0, I3 / I2]
        conic_type = CONIC_TYPE_INTERSECT_LINE if I3 == 0 else CONIC_TYPE_HYPERBOLA
        return [conic_type, corr]


def intersect_line_conic(x1, y1, x2, y2, a11, a12, a22, a1, a2, a0):
    vx, vy = x2 - x1, y2 - y1
    at2 = a11 * vx ** 2 + 2 * a12 * vx * vy + a22 * vy ** 2
    f1x1y1 = a11 * x1 + a12 * y1 + a1
    f2x1y1 = a12 * x1 + a22 * y1 + a2
    at1 = 2 * (f1x1y1 * vx + f2x1y1 * vy)
    at0 = a11 * x1 ** 2 + 2 * a12 * x1 * y1 + a22 * y1 ** 2 + 2 * a1 * x1 + 2 * a2 * y1 + a0

    roots = real_root_tow_power_equation(at2, at1, at0)

    if roots is None:
        return None

    def get_xy_by_t(t):
        return [x1 + t * vx, y1 + t * vy]

    if len(roots) == 1:
        return [get_xy_by_t(roots[0])]

    return get_xy_by_t(roots[0]), get_xy_by_t(roots[1])


def get_conic_symmetry_center(a11, a12, a22, a1, a2, a0):
    mat = [
        [a11, a12, -a1],
        [a12, a22, -a2]
    ]
    mat2 = rref(mat)
    if mat2[1][1] != 0:
        return mat2[0][2], mat2[1][2]
    return None


def get_conic_asymptote_vector(a11, a12, a22, a1, a2, a0):
    roots = real_root_tow_power_equation(a22, 2 * a12, a11)

    return roots


def get_diameter_direction_by_conic(a11, a12, a22, a1, a2, a0):
    mat = [
        [a11, a12],
        [a12, a22]
    ]

    return eig_n(mat)


def get_coefficient_by_conic(conic):
    conic_type = conic.conic_type_cls_var
    if conic_type is None:
        coe = conic.basic_data
    elif conic_type == CONIC_TYPE_CIRCLE:
        coe = get_circle_equation_coefficient_by_center_radius(*conic.basic_data[0], conic.basic_data[1])
    elif conic_type == CONIC_TYPE_ELLIPSE or conic_type == CONIC_TYPE_HYPERBOLA:
        x1, y1 = conic.basic_data[0]
        x2, y2 = conic.basic_data[1]
        d = conic.basic_data[2]
        coe = get_central_conic_equation_coefficient_by_focus_d(x1, y1, x2, y2, d)
    elif conic_type == CONIC_TYPE_PARABOLA:
        xp, yp = conic.basic_data[0]
        xv, yv = conic.basic_data[1]
        x2, y2 = get_rotate_point_by_two_point(xv, yv, xp, yp, pi_half)
        coe = get_parabola_equation_coefficient_by_focus_line(xp, yp, xv, yv, x2, y2)
    else:
        coe = None
    return coe


def conic_tangent_vector(x0, y0, a11, a12, a22, a1, a2, a0):
    fx0y0 = get_conic_value_by_point(x0, y0, a11, a12, a22, a1, a2, a0)
    f1x0y0 = a11 * x0 + a12 * y0 + a1
    f2x0y0 = a12 * x0 + a22 * y0 + a2
    if is_zero(fx0y0):
        x = -f2x0y0
        y = f1x0y0
        return [[x, y]]

    def fi(u, v):
        return a11 * u ** 2 + 2 * a12 * u * v + a22 * v ** 2

    cu = f1x0y0 ** 2 - fx0y0 * a11
    cuv = f1x0y0 * f2x0y0 - fx0y0 * a12
    cv = f2x0y0 ** 2 - fx0y0 * a22
    res = []
    if not is_zero(cv):
        roots = real_root_tow_power_equation(cv, 2 * cuv, cu)
        if roots is None:
            return
        if len(roots) == 1:
            fii = fi(1, roots[0])
            if not is_zero(fii):
                res.append([1, roots[0], -(f1x0y0 + roots[0] * f2x0y0) / fii])
                return res
        fii1 = fi(1, roots[0])
        if not is_zero(fii1):
            res.append([1, roots[0], -(f1x0y0 + roots[0] * f2x0y0) / fii1])
        fii2 = fi(1, roots[1])
        if not is_zero(fii2):
            res.append([1, roots[1], -(f1x0y0 + roots[1] * f2x0y0) / fii2])
        return res
    if not is_zero(cuv):
        fii1 = fi(0, 1)
        if not is_zero(fii1):
            res.append([0, 1, -f2x0y0 / fii1])
        fii2 = fi(-2 * cuv, cu)
        if not is_zero(fii2):
            res.append([-2 * cuv, cu, -(-2 * cuv * f1x0y0 + cu * f2x0y0) / fii2])
        return res
    return



from ..core.king_geometry import KingGeometry
from matplotlib.pyplot import Line2D
import abc
from matplotlib.transforms import Affine2D
from matplotlib.text import Text


class KingCurve(KingGeometry, Line2D, metaclass=abc.ABCMeta):
    def __init__(self, depends=(), essential_data=None, create_type=1, infinite=False):
        self.text = Text(0, 0, '')
        super(KingGeometry, self).__init__([], [])
        super(KingCurve, self).__init__(depends=depends, essential_data=essential_data,
                                        create_type=create_type, infinite=infinite)
        self.text.set_text(self.get_label())
        self.mouse_select_priority = 2
        if create_type == 1:
            self.mouse_translation = True

    def set_visible(self, b):
        super(KingCurve, self).set_visible(b)
        if not b:
            self.mouse_translation = False

    def set_figure(self, figure):
        self.text.set_figure(figure)
        super(KingCurve, self).set_figure(figure)

    def set_axes(self, axes):
        self.text.set_axes(axes)
        super().set_axes(axes)

    def set_transform(self, transform):
        self.text.set_transform(transform + Affine2D().translate(2, 2))
        super().set_transform(transform)

    def set_data(self, x, y):
        if isinstance(x, list) and len(x) > 0:
            len_x = len(x)
            if len_x == 2:
                pos = ((x[0] + x[1]) / 2, (y[0] + y[1]) / 2)
            else:
                pos = (x[-1], y[-1])
            self.text.set_position(pos)
        else:
            self.text.set_position((x, y))
        super().set_data(x, y)

    def draw(self, renderer):
        super().draw(renderer)
        self.text.draw(renderer)

    @abc.abstractmethod
    def generate_xy_data_from_basic_data(self):
        pass

    def assign_basic_data_to_artist(self):
        self.set_data(*self.generate_xy_data_from_basic_data())



from .king_curve import KingCurve
from ..core.king_figure import KingFigure
from ..alg.conic_alg import *
from ..alg.plane_geometry_alg import get_edge_two_point_by_line
from ..alg.basic_alg import nan
from ..alg.circle_alg import get_circle_xy_data_by_center_radius
from ..alg.ellipse_alg import get_ellipse_xy_data_by_standard_equation
from ..alg.hyperbola_alg import get_hyperbola_xy_data_by_standard_equation
from ..alg.parabola_alg import get_parabola_xy_data_by_standard_equation


class KingConic(KingCurve):
    conic_type_cls_var = None

    def __init__(self, depends=(), essential_data=None, create_type=1, infinite=True):
        self.conic_type = None
        self.conic_basic_data = None
        super(KingConic, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                        infinite=infinite)

    def generate_xy_data_from_basic_data(self):
        x_min, x_max = KingFigure().get_x_lim()
        y_min, y_max = KingFigure().get_y_lim()
        t1, t2, translation_coefficient = get_simplest_conic_equation_and_transform_by_conic_equation(*self.basic_data)
        type_basic_data = get_conic_type_and_basic_data_by_simplest_equation(*translation_coefficient)
        kind = type_basic_data[0]
        self.conic_type = kind
        if kind == CONIC_TYPE_CAN_NOT_PLOT:
            self.conic_basic_data = [[], []]
            return self.conic_basic_data
        if kind == CONIC_TYPE_POINT:
            self.conic_basic_data = t1(0, 0)
            return self.conic_basic_data
        if kind == CONIC_TYPE_LINE:
            x1, y1 = t1(*type_basic_data[1][0])
            x2, y2 = t1(*type_basic_data[1][1])
            basic_data = get_edge_two_point_by_line(x1, y1, x2, y2, x_min, x_max, y_min, y_max)
            self.conic_basic_data = basic_data
            return [[basic_data[0][0], basic_data[1][0]], [basic_data[0][1], basic_data[1][1]]]
        if kind == CONIC_TYPE_PARALLEL_LINE or kind == CONIC_TYPE_INTERSECT_LINE:
            x1, y1 = type_basic_data[1][0][0]
            x2, y2 = type_basic_data[1][0][1]
            x3, y3 = type_basic_data[1][1][0]
            x4, y4 = type_basic_data[1][1][1]
            x1, y1 = t1(x1, y1)
            x2, y2 = t1(x2, y2)
            x3, y3 = t1(x3, y3)
            x4, y4 = t1(x4, y4)
            l1 = get_edge_two_point_by_line(x1, y1, x2, y2, x_min, x_max, y_min, y_max)
            l2 = get_edge_two_point_by_line(x3, y3, x4, y4, x_min, x_max, y_min, y_max)
            self.conic_basic_data = [l1, l2]
            return [[l1[0][0], l1[1][0], nan, l2[0][0], l2[1][0]], [l1[0][1], l1[1][1], nan, l2[0][1], l2[1][1]]]
        if kind == CONIC_TYPE_CIRCLE:
            o = t1(0, 0)
            r = type_basic_data[1]
            self.conic_basic_data = [o, r]

            return get_circle_xy_data_by_center_radius(*o, r)
        if kind == CONIC_TYPE_ELLIPSE:
            standard_xy = \
                get_ellipse_xy_data_by_standard_equation(-translation_coefficient[0] / translation_coefficient[5],
                                                         0,
                                                         -translation_coefficient[2] / translation_coefficient[5],
                                                         0, 0, -1)
            self.conic_basic_data = [t1(*type_basic_data[1][0]), t1(*type_basic_data[1][1]), type_basic_data[1][2]]

            return t1(*standard_xy)
        if kind == CONIC_TYPE_HYPERBOLA:
            center = t2((x_min + x_max) / 2, (y_min + y_max) / 2)
            x_lim = 0.71 * (x_max - x_min)
            y_lim = 0.71 * (y_max - y_min)

            standard_xy = \
                get_hyperbola_xy_data_by_standard_equation(translation_coefficient[0] / translation_coefficient[5],
                                                           0,
                                                           translation_coefficient[2] / translation_coefficient[5],
                                                           0, 0, 1,
                                                           center[0] - x_lim,
                                                           center[0] + x_lim,
                                                           center[1] - y_lim,
                                                           center[1] + y_lim)
            self.conic_basic_data = [t1(*type_basic_data[1][0]), t1(*type_basic_data[1][1]), type_basic_data[1][2]]

            return t1(*standard_xy)
        if kind == CONIC_TYPE_PARABOLA:
            center = t2((x_min + x_max) / 2, (y_min + y_max) / 2)
            x_lim = 0.71 * (x_max - x_min)
            y_lim = 0.71 * (y_max - y_min)

            standard_xy = get_parabola_xy_data_by_standard_equation(*translation_coefficient,
                                                                    center[0] - x_lim,
                                                                    center[0] + x_lim,
                                                                    center[1] - y_lim,
                                                                    center[1] + y_lim)
            self.conic_basic_data = [t1(*type_basic_data[1][0]), t1(*type_basic_data[1][1])]

            return t1(*standard_xy)



from .king_curve import KingCurve
from ..alg.basic_alg import linspace


class KingParametricEquation(KingCurve):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        """
        {'x': x, 'y': y, 't_min': min, 't_max': max, 'count': c}
        """
        super(KingParametricEquation, self).__init__(depends=depends, essential_data=essential_data,
                                                     create_type=create_type)

    def generate_basic_data(self, depends=(), essential_data=None):
        return essential_data

    def generate_xy_data_from_basic_data(self):
        t_min = self.basic_data['t_min']
        t_max = self.basic_data['t_max']
        count = self.basic_data['count']
        x = self.basic_data['x']
        y = self.basic_data['y']
        if count is None:
            count = 100

        t = linspace(t_min, t_max, count)
        return [x(t), y(t)]



from .king_conic_graph import *
from ..alg.ellipse_alg import get_ellipse_basic_data_by_three_point, get_ellipse_basic_data_by_isosceles_triangle
from ..alg.plane_geometry_alg import get_distance_by_tow_point
from ..core.king_decorater import geometry_generate
from ..alg.plane_geometry_alg import get_vertical_point_by_point_line
from ..alg.basic_alg import *


@geometry_generate(cls=KingCircle)
def circle_center_edge_point(depends=(), essential_data=None):
    o_x = depends[0].basic_data[0]
    o_y = depends[0].basic_data[1]
    c_x = depends[1].basic_data[0]
    c_y = depends[1].basic_data[1]
    return [[o_x, o_y], get_distance_by_tow_point(o_x, o_y, c_x, c_y)]


@geometry_generate(cls=KingCircle)
def circle_three_point(depends=(), essential_data=None):
    mat = [[2 * p.basic_data[0], 2 * p.basic_data[1], 1, -(p.basic_data[0] ** 2 + p.basic_data[1] ** 2)]
           for p in depends]
    mat2 = rref(mat)
    if mat2[2][2] == 1:
        D = mat2[0][3]
        E = mat2[1][3]
        F = mat2[2][3]
        return [[-D, -E], sqrt(D ** 2 + E ** 2 - F)]
    return [[nan, nan], nan]


@geometry_generate(cls=KingEllipse)
def ellipse_three_point(depends=(), essential_data=None):
    x1, y1 = depends[0].basic_data
    x2, y2 = depends[1].basic_data
    x3, y3 = depends[2].basic_data

    return get_ellipse_basic_data_by_three_point(x1, y1, x2, y2, x3, y3)


@geometry_generate(cls=KingEllipse)
def ellipse_isosceles_triangle(depends=(), essential_data=None):
    p1x, p1y = depends[0].basic_data
    p2x, p2y = depends[1].basic_data
    b = essential_data[0]

    return get_ellipse_basic_data_by_isosceles_triangle(p1x, p1y, p2x, p2y, b)


@geometry_generate(cls=KingHyperbola)
def hyperbola_three_point(depends=(), essential_data=None):
    x1, y1 = depends[0].basic_data
    x2, y2 = depends[1].basic_data
    x3, y3 = depends[2].basic_data
    b2 = get_distance_by_tow_point(x1, y1, x3, y3) - get_distance_by_tow_point(x2, y2, x3, y3)

    return [depends[0].basic_data, depends[1].basic_data, b2]


@geometry_generate(cls=KingParabola)
def parabola_point_line(depends=(), essential_data=None):
    xp, yp = depends[0].basic_data
    x1, y1 = depends[1].basic_data[0]
    x2, y2 = depends[1].basic_data[1]
    xv, yv = get_vertical_point_by_point_line(xp, yp, x1, y1, x2, y2)
    return [[xp, yp], [xv, yv]]



from .basic_alg import linspace, sqrt, nan
from .plane_geometry_alg import get_transform_vector_to_y_up


def get_hyperbola_focus_diff_by_standard_equation(a11, a12, a22, a1, a2, a0):
    """
    a11 = (+/-) 1/a^2
    a22 = (+-) 1/b^2
    a12 = a1 = a2 = 0
    a0 = (+-) 1
    """
    if a11 < 0:
        a11 = -a11
        a22 = -a22
        a0 = -a0
    if a0 > 0:
        a11_1 = 1 / a11
        cc = a11_1 - 1 / a22
        a = sqrt(a11_1)
        c = sqrt(cc)
        return [[c, 0], [-c, 0], 2 * a]
    a22_1 = 1 / a22
    cc = 1 / a11 - a22_1
    b = sqrt(-a22_1)
    c = sqrt(cc)
    return [[0, c], [0, -c], 2 * b]


def get_hyperbola_equation_coefficient_by_simplest_focus_diff(x1, y1, x2, y2, diff):
    """
    [1/a^2, 0, -1/b^2, 0, 0, (+-) 1]
    """
    if y1 == 0:
        c = abs(x1)
        cc = c ** 2
        a = diff / 2
        aa = a ** 2
        bb = cc - aa
        return [1 / aa, 0, -1 / bb, 0, 0, -1]
    c = abs(y1)
    cc = c ** 2
    b = diff / 2
    bb = b ** 2
    aa = cc - bb
    return [1 / aa, 0, -1 / bb, 0, 0, 1]


def get_hyperbola_xy_data_by_standard_equation(a11, a12, a22, a1, a2, a0, x_min, x_max, y_min, y_max):
    if a11 < 0:
        a11 = -a11
        a22 = -a22
        a0 = -a0
    if a0 > 0:
        x = linspace(x_min, x_max, 300)
        b = 1 / sqrt(-a22)
        y = b * sqrt(1 + x ** 2 * a11)
        result_x = []
        result_y = []
        result_x.extend(x)
        result_x.append(nan)
        result_x.extend(x)
        result_y.extend(y)
        result_y.append(nan)
        result_y.extend(-y)
        return [result_x, result_y]
    y = linspace(y_min, y_max, 300)
    a = 1 / sqrt(a11)
    x = a * sqrt(1 - y ** 2 * a22)
    result_x = []
    result_y = []
    result_x.extend(x)
    result_x.append(nan)
    result_x.extend(-x)
    result_y.extend(y)
    result_y.append(nan)
    result_y.extend(y)
    return [result_x, result_y]


def get_hyperbola_xy_data_by_simplest_focus_diff(x1, y1, x2, y2, diff, x_min, x_max, y_min, y_max):
    return get_hyperbola_xy_data_by_standard_equation(
        *get_hyperbola_equation_coefficient_by_simplest_focus_diff(x1, y1, x2, y2, diff), x_min, x_max, y_min, y_max)


def get_hyperbola_xy_data_by_focus_diff(x1, y1, x2, y2, diff, x_min, x_max, y_min, y_max):
    """
    get hyperbola point set by two focus, difference, x limit and y limit.
    return : tow set represent for tow line.
    """
    b = abs(diff) / 2
    b_2 = b ** 2
    c2_2 = (x1 - x2) ** 2 + (y1 - y2) ** 2
    c_2 = c2_2 / 4
    a_2 = c_2 - b_2
    if c_2 <= b_2:
        raise Exception("Can not plot a hyperbola")

    def line_func(x):
        return b * sqrt(1 + x ** 2 / a_2)

    center_x = (x_min + x_max) / 2
    center_y = (y_min + y_max) / 2
    x_lim = x_max - x_min
    t1, t2 = get_transform_vector_to_y_up(x1, y1, x2, y2)
    center_x_new, center_y_new = t2(center_x, center_y)
    x_lim_new = 0.71 * x_lim
    x_set = linspace(center_x_new - x_lim_new, center_x_new + x_lim_new, 300)
    y_set_1 = line_func(x_set)
    y_set_2 = -y_set_1
    result_1 = t1(x_set, y_set_1)
    result_2 = t1(x_set, y_set_2)

    x_data = []
    x_data.extend(result_1[0])
    x_data.append(nan)
    x_data.extend(result_2[0])
    y_data = []
    y_data.extend(result_1[1])
    y_data.append(nan)
    y_data.extend(result_2[1])

    return [x_data, y_data]



from .basic_alg import unit_circle


def get_circle_equation_coefficient_by_center_radius(x1, y1, r):
    return [1, 0, 1, -x1, -y1, x1 ** 2 + y1 ** 2 - r ** 2]


def get_circle_xy_data_by_center_radius(x1, y1, r):
    x_data = r * unit_circle[0]
    y_data = r * unit_circle[1]
    x_data = x_data + x1
    y_data = y_data + y1
    return x_data, y_data



from .king_plane_geometry_graph import *
from ..alg.plane_geometry_alg import get_middle_point_by_two_point, get_rotate_point_by_two_point, \
    get_vertical_point_by_point_line, get_symmetric_point_by_point_line, get_point_by_matrix, \
    get_intersect_point_corr_by_four_point, get_symmetric_point_by_point, \
    get_intersect_point_corr_by_four_point_segment_and_segment, \
    get_intersect_point_corr_by_four_point_line_and_segment, \
    get_intersect_point_corr_by_circle_and_circle
from ..core.king_decorater import geometry_generate
from ..core.king_decorater_multi_geometry import geometries_generate
from ..core.king_figure import KingFigure
from math import inf
from ..alg.basic_alg import nan, pi_half, tow_empty_line_basic_data, empty_line_basic_data, empty_point_basic_data
from ..alg.conic_alg import intersect_line_conic, get_central_conic_equation_coefficient_by_focus_d, \
    get_coefficient_by_conic
from ..alg.conic_alg import get_conic_symmetry_center, get_conic_asymptote_vector, get_diameter_direction_by_conic, \
    conic_tangent_vector
from ..geometry.king_conic import KingConic
from ..geometry.king_conic_graph import KingHyperbola


@geometry_generate(cls=KingPoint)
def point_middle(depends=(), essential_data=None):
    return get_middle_point_by_two_point(*depends[0].basic_data, *depends[1].basic_data)


@geometry_generate(cls=KingPoint)
def point_rotate(depends=(), essential_data=None):
    return get_rotate_point_by_two_point(depends[0].basic_data[0], depends[0].basic_data[1],
                                         depends[1].basic_data[0], depends[1].basic_data[1],
                                         essential_data[0])


@geometry_generate(cls=KingPoint)
def point_vertical_point_line(depends=(), essential_data=None):
    return get_vertical_point_by_point_line(*depends[0].basic_data,
                                            *depends[1].basic_data[0],
                                            *depends[1].basic_data[1])


@geometry_generate(cls=KingPoint)
def point_symmetric_point_line(depends=(), essential_data=None):
    return get_symmetric_point_by_point_line(*depends[0].basic_data, *depends[1].basic_data[0],
                                             *depends[1].basic_data[1])


@geometry_generate(cls=KingPoint)
def point_symmetric_point(depends=(), essential_data=None):
    return get_symmetric_point_by_point(*depends[0].basic_data, *depends[1].basic_data)


@geometry_generate(cls=KingPoint)
def point_matrix(depends=(), essential_data=None):
    x1, y1 = depends[0].basic_data
    mat = essential_data[0]
    return get_point_by_matrix(mat, x1, y1)


@geometry_generate(cls=KingPoint)
def point_intersect_by_two_line(depends=(), essential_data=None):
    l1 = depends[0]
    l2 = depends[1]
    p1 = l1.basic_data[0]
    p2 = l1.basic_data[1]
    p3 = l2.basic_data[0]
    p4 = l2.basic_data[1]

    return get_intersect_point_corr_by_four_point(*p1, *p2, *p3, *p4)


@geometry_generate(cls=KingPoint)
def point_intersect_by_segment_and_segment(depends=(), essential_data=None):
    l1 = depends[0]
    l2 = depends[1]
    p1 = l1.basic_data[0]
    p2 = l1.basic_data[1]
    p3 = l2.basic_data[0]
    p4 = l2.basic_data[1]

    return get_intersect_point_corr_by_four_point_segment_and_segment(*p1, *p2, *p3, *p4)


@geometry_generate(cls=KingPoint)
def point_intersect_by_line_and_segment(depends=(), essential_data=None):
    l1 = depends[0]
    l2 = depends[1]
    p1 = l1.basic_data[0]
    p2 = l1.basic_data[1]
    p3 = l2.basic_data[0]
    p4 = l2.basic_data[1]

    return get_intersect_point_corr_by_four_point_line_and_segment(*p1, *p2, *p3, *p4)


@geometries_generate(classes=[KingPoint], numbers=[2])
def point_intersect_by_circle_and_circle(depends=(), essential_data=None):
    c1 = depends[0]
    c2 = depends[1]

    return get_intersect_point_corr_by_circle_and_circle(c1.basic_data[0][0], c1.basic_data[0][1], c1.basic_data[1],
                                                         c2.basic_data[0][0], c2.basic_data[0][1], c2.basic_data[1])


@geometry_generate(cls=KingPoint)
def point_conic_symmetry_center(depends=(), essential_data=None):
    center = get_conic_symmetry_center(*depends[0].basic_data)
    if center is None:
        return [nan, nan]
    return center


@geometry_generate(cls=KingLine)
def line_kb(depends=(), essential_data=None):
    k = essential_data[0]
    b = essential_data[1]
    if k == inf or k == -inf:
        y_min, y_max = KingFigure().get_y_lim()
        return [[b, y_min], [b, y_max]]

    x_min, x_max = KingFigure().get_x_lim()

    return [[x_min, k * x_min + b], [x_max, k * x_max + b]]


@geometry_generate(cls=KingLine)
def line_point_k(depends=(), essential_data=None):
    k = essential_data[0]
    if k == inf:
        y_min, y_max = KingFigure().get_y_lim()
        return [[depends[0].basic_data[0], y_min], [depends[0].basic_data[0], y_max]]
    x_min, x_max = KingFigure().get_x_lim()
    x1, y1 = depends[0].basic_data
    y1_k_x1 = y1 - k * x1
    return [[x_min, k * x_min + y1_k_x1], [x_max, k * x_max + y1_k_x1]]


@geometry_generate(cls=KingLine)
def line_general(depends=(), essential_data=None):
    a = essential_data[0]
    b = essential_data[1]
    c = essential_data[2]
    if a == 0:
        if b == 0:
            if c != 0:
                raise Exception("a=0,b=0,c!=0")
            return [[0, 0], [0, 0]]
        x_min, x_max = KingFigure().get_x_lim()
        _c_b = -c / b
        return [[x_min, _c_b], [x_max, _c_b]]
    if b == 0:
        if c == 0:
            return [[0, 0], [0, 0]]
        y_min, y_max = KingFigure().get_y_lim()
        _a_c = -a / c
        return [[_a_c, y_min], [_a_c, y_max]]
    x_min, x_max = KingFigure().get_x_lim()
    _a_b = -a / b
    _c_b = -c / b
    return [[x_min, _a_b * x_min + _c_b], [x_max, _a_b * x_max + _c_b]]


@geometries_generate(classes=[KingPoint], numbers=[2])
def point_intersect_line_conic(depends=(), essential_data=None):
    line = depends[0]
    ps = intersect_line_conic(*line.basic_data[0], *line.basic_data[1], *get_coefficient_by_conic(depends[1]))
    if ps is None:
        return [[nan] * 2] * 2
    if len(ps) == 1:
        return [ps[0], [nan, nan]]
    return [[ps[0][0], ps[0][1]], [ps[1][0], ps[1][1]]]


@geometries_generate(classes=[KingLine], numbers=[2])
def conic_asymptote(depends=(), essential_data=None):
    conic = depends[0]
    if isinstance(conic, KingConic):
        a11, a12, a22, a1, a2, a0 = conic.basic_data
    elif isinstance(conic, KingHyperbola):
        x1, y1 = conic.basic_data[0]
        x2, y2 = conic.basic_data[1]
        d = conic.basic_data[2]
        a11, a12, a22, a1, a2, a0 = get_central_conic_equation_coefficient_by_focus_d(x1, y1, x2, y2, d)
    else:
        return tow_empty_line_basic_data
    vec = get_conic_asymptote_vector(a11, a12, a22, a1, a2, a0)
    center = get_conic_symmetry_center(a11, a12, a22, a1, a2, a0)
    if center is None:
        return tow_empty_line_basic_data
    if vec is None:
        return tow_empty_line_basic_data
    return [get_edge_two_point_by_line(center[0], center[1], center[0] + 1, center[1] + vec[0],
                                       *KingFigure().get_x_lim(), *KingFigure().get_y_lim()),
            get_edge_two_point_by_line(center[0], center[1], center[0] + 1, center[1] + vec[1],
                                       *KingFigure().get_x_lim(), *KingFigure().get_y_lim())]


@geometries_generate(classes=[KingLine], numbers=[2])
def conic_symmetry_lines(depends=(), essential_data=None):
    conic = depends[0]
    a11, a12, a22, a1, a2, a0 = conic.basic_data
    vec = get_diameter_direction_by_conic(a11, a12, a22, a1, a2, a0)
    center = get_conic_symmetry_center(a11, a12, a22, a1, a2, a0)
    if center is None:
        center = conic.conic_basic_data[0]
        n = 0 if vec[2][0] == 0 else 1
        return [get_edge_two_point_by_line(center[0], center[1], center[0] + vec[n][0], center[1] + vec[n][1],
                                           *KingFigure().get_x_lim(), *KingFigure().get_y_lim()),
                empty_line_basic_data]
    return [get_edge_two_point_by_line(center[0], center[1], center[0] + vec[0][0], center[1] + vec[0][1],
                                       *KingFigure().get_x_lim(), *KingFigure().get_y_lim()),
            get_edge_two_point_by_line(center[0], center[1], center[0] + vec[1][0], center[1] + vec[1][1],
                                       *KingFigure().get_x_lim(), *KingFigure().get_y_lim())]


@geometries_generate(classes=[KingLine, KingPoint], numbers=[2, 2])
def tangent_line_by_conic(depends=(), essential_data=None):
    p = depends[0].basic_data
    vs = conic_tangent_vector(*p, *get_coefficient_by_conic(depends[1]))
    if vs is None:
        return empty_line_basic_data, empty_line_basic_data, empty_point_basic_data, empty_point_basic_data
    if len(vs) == 1:
        l1 = get_edge_two_point_by_line(p[0], p[1], p[0] + vs[0][0], p[1] + vs[0][1], *KingFigure().get_x_lim(),
                                        *KingFigure().get_y_lim())
        l2 = empty_line_basic_data
        if len(vs[0]) == 2:
            p1 = empty_point_basic_data
            p2 = empty_point_basic_data
        else:
            p1 = [p[0] + vs[0][2] * vs[0][0], p[1] + vs[0][2] * vs[0][1]]
            p2 = empty_point_basic_data
        return [l1, l2, p1, p2]

    l1 = get_edge_two_point_by_line(p[0], p[1], p[0] + vs[0][0], p[1] + vs[0][1], *KingFigure().get_x_lim(),
                                    *KingFigure().get_y_lim())
    l2 = get_edge_two_point_by_line(p[0], p[1], p[0] + vs[1][0], p[1] + vs[1][1], *KingFigure().get_x_lim(),
                                    *KingFigure().get_y_lim())
    p1 = [p[0] + vs[0][2] * vs[0][0], p[1] + vs[0][2] * vs[0][1]]
    p2 = [p[0] + vs[1][2] * vs[1][0], p[1] + vs[1][2] * vs[1][1]]

    return [l1, l2, p1, p2]



"""
support numeric calculate and symbolic calculate.
can choose which type of function to invoke automatically.
symbolic calculate must pass sympy object and return sympy object.
"""

import numpy as np
import ctypes
import sympy
import platform
import random

pi = np.pi
pi_half = pi / 2
pi_one_third = pi / 3
pi_quarter = pi / 4
one_degree = pi / 180
pi_2 = pi * 2
theta_list = np.linspace(0, pi_2, 360)
theta_list.flags.writeable = False
unit_circle = (np.cos(theta_list), np.sin(theta_list))
unit_circle[0].flags.writeable = False
unit_circle[1].flags.writeable = False
nan = np.nan
libc = None
sys_type = platform.system()
tow_empty_line_basic_data = [[[nan] * 2] * 2] * 2
empty_line_basic_data = [[nan] * 2] * 2
empty_point_basic_data = [nan, nan]
accuracy = 0.1e-14

try:
    lib_location = None
    if sys_type == "Windows":
        lib_location = '../../../Clib/bin/libKingAlg.dll'
    elif sys_type == "Linux":
        lib_location = '../../../Clib/bin/libKingAlg.so'
    libc = ctypes.cdll.LoadLibrary(lib_location)
except OSError:
    libc = None
    print(OSError)


def is_zero(n):
    return abs(n) <= accuracy


def greater_than_zero(n):
    return n > accuracy


def less_than_zero(n):
    return n < -accuracy


def transfer_sympy_matrix_to_array(matrix):
    row = matrix.rows
    col = matrix.cols
    result = []

    for i in range(row):
        result.append([])
        for j in range(col):
            result[i].append(matrix[i * col + j])

    return result


def array(p_object):
    return np.array(p_object)


def linspace(start, stop, num=50, endpoint=True, retstep=False, dtype=None, axis=0):
    return np.linspace(start, stop, num=num, endpoint=endpoint, retstep=retstep, dtype=dtype, axis=axis)


def is_sympy_object(o):
    if isinstance(o, sympy.Symbol):
        return True
    if isinstance(o, sympy.Integer):
        return True
    if isinstance(o, sympy.Matrix):
        return True
    if isinstance(o, sympy.Rational):
        return True
    return False


def cos_n(t):
    return np.cos(t)


def cos_s(t):
    return sympy.cos(t)


def cos(t):
    if is_sympy_object(t):
        return cos_s(t)
    return cos_n(t)


def arccos_n(t):
    return np.arccos(t)


def arccos_s(t):
    return sympy.acos(t)


def arccos(t):
    if is_sympy_object(t):
        return arccos_s(t)
    return arccos_n(t)


def sin_n(t):
    return np.sin(t)


def sin_s(t):
    return sympy.sin(t)


def sin(t):
    if is_sympy_object(t):
        return sin_s(t)
    return sin_n(t)


def arctan2_n(x1, x2):
    return np.arctan2(x1, x2)


def arctan2_s(x1, x2):
    return sympy.atan2(x1, x2)


def arctan2(x1, x2):
    if is_sympy_object(x1) or is_sympy_object(x2):
        return arctan2_s(x1, x2)
    return arctan2_n(x1, x2)


def sqrt_n(x):
    return np.sqrt(x)


def sqrt_s(x):
    return sympy.sqrt(x)


def sqrt(x):
    if is_sympy_object(x):
        return sqrt_s(x)
    return sqrt_n(x)


def matmul_n(x1, x2):
    return np.matmul(x1, x2)


def matmul_s(x1, x2):
    if not is_sympy_object(x1):
        x1 = sympy.Matrix(x1)
    if not is_sympy_object(x2):
        x2 = sympy.Matrix(x2)
    return x1 * x2


def matmul(x1, x2):
    if is_sympy_object(x1) or is_sympy_object(x2):
        return matmul_s(x1, x2)
    return matmul_n(x1, x2)


def rref_s(mat):
    return mat.rref(pivots=False)


def rref_n(mat):
    c_rref = libc.rref
    c_rref.restype = None

    row = len(mat)
    col = len(mat[0])

    result = ((ctypes.c_double * col) * row)()

    for i in range(row):
        for j in range(col):
            result[i][j] = ctypes.c_double(mat[i][j])

    c_rref(row, col, result)

    return result


def rref(mat):
    if is_sympy_object(mat):
        return rref_s(mat)
    if libc is None:
        sympy_mat = rref_s(sympy.Matrix(mat))
        arr = transfer_sympy_matrix_to_array(sympy_mat)
        for i in range(len(arr)):
            for j in range(len(arr[0])):
                arr[i][j] = float(arr[i][j])
        return arr
    return rref_n(mat)


def get_main_element_index_list_s(row_simplest_matrix):
    result = []
    row = row_simplest_matrix.rows
    col = row_simplest_matrix.cols
    i = 0
    j = 0
    while i < row:
        while j < col:
            if row_simplest_matrix[i * col + j] != 0:
                result.append(j)
                j += 1
                break
            j += 1
        if j == col:
            return result
        i += 1
    return result


def get_main_element_index_list_n(row_simplest_matrix):
    result = []
    row = len(row_simplest_matrix)
    col = len(row_simplest_matrix[0])
    i = 0
    j = 0
    while i < row:
        while j < col:
            if row_simplest_matrix[i][j] != 0:
                result.append(j)
                j += 1
                break
            j += 1
        if j == col:
            return result
        i += 1
    return result


def random_color_number():
    n = hex(random.randint(0, 255))[2:]
    if len(n) == 1:
        return '0' + n
    return n


def random_color():
    return '#' + random_color_number() + random_color_number() + random_color_number()


def get_main_element_index_list(row_simplest_matrix):
    if is_sympy_object(row_simplest_matrix):
        return get_main_element_index_list_s(row_simplest_matrix)
    return get_main_element_index_list_n(row_simplest_matrix)


def eig_n(mat):
    m = np.array(mat)
    eig = np.linalg.eig(m)

    return [[eig[1][0][0], eig[1][1][0]], [eig[1][0][1], eig[1][1][1]], [eig[0][0], eig[0][1]]]


def det(mat):
    if is_sympy_object(mat):
        return sympy.det(mat)
    if not isinstance(mat, np.ndarray):
        mat = np.array(mat)
    return np.linalg.det(mat)


def trace(mat):
    if is_sympy_object(mat):
        return mat.trace()
    if not isinstance(mat, np.ndarray):
        mat = np.array(mat)
    return mat.trace()


def real_root_tow_power_equation(a, b, c):
    """
    ax^2+bx+c=0
    """
    delta = b ** 2 - 4 * a * c
    if abs(delta) < 0.1e-7:
        delta = 0
    if delta < 0:
        return None
    if delta == 0:
        return [-b / (2 * a)]
    sqrt_delta = sqrt(delta)
    a2 = 2 * a
    a2_b = -b / a2
    return [a2_b - sqrt_delta / a2, a2_b + sqrt_delta / a2]


def is_same_negative_or_positive(x1, x2):
    if x1 > 0 and x2 > 0:
        return 1
    if x1 < 0 and x2 < 0:
        return -1
    if x1 == 0 and x2 == 0:
        return 0
    return None


def is_object_in_class_list(o, cls_list):
    for i in cls_list:
        if isinstance(o, i):
            return True
    return False



from .basic_alg import sqrt, unit_circle
from .plane_geometry_alg import get_transform_vector_to_y_up, get_distance_by_tow_point


def get_ellipse_focus_add_by_standard_equation(a11, a12, a22, a1, a2, a0):
    """
    a11 = 1/a^2
    a22 = 1/b^2
    a12 = a1 = a2 = 0
    a0 = -1
    """
    if a11 < a22:
        a11_1 = 1 / a11
        c = sqrt(a11_1 - 1 / a22)
        return [[c, 0], [-c, 0], 2 * sqrt(a11_1)]
    a22_1 = 1 / a22
    c = sqrt(a22_1 - 1 / a11)
    return [[0, c], [0, -c], 2 * sqrt(a22_1)]


def get_ellipse_equation_coefficient_by_simplest_focus_add(x1, y1, x2, y2, add):
    """
    [1/a^2, 0, 1/b^2, 0, 0, -1]
    """
    if y1 == 0:
        a = add / 2
        aa = a * a
        c = abs(x1)
        bb = aa - c ** 2
        return [1 / aa, 0, 1 / bb, 0, 0, -1]
    b = add / 2
    bb = b * b
    c = abs(y1)
    aa = bb - c ** 2
    return [1 / aa, 0, 1 / bb, 0, 0, -1]


def get_ellipse_xy_data_by_standard_equation(a11, a12, a22, a1, a2, a0):
    a = 1 / sqrt(a11)
    b = 1 / sqrt(a22)
    return [unit_circle[0] * a, unit_circle[1] * b]


def get_ellipse_xy_data_by_simplest_focus_add(x1, y1, x2, y2, add):
    return get_ellipse_xy_data_by_standard_equation(
        *get_ellipse_equation_coefficient_by_simplest_focus_add(x1, y1, x2, y2, add))


def get_ellipse_xy_data_by_focus_add(x1, y1, x2, y2, add):
    b = add / 2
    dd = (x1 - x2) ** 2 + (y1 - y2) ** 2
    cc = dd / 4
    bb = b ** 2
    aa = bb - cc
    a = sqrt(aa)
    x_data = unit_circle[0] * a
    y_data = unit_circle[1] * b
    t1, t2 = get_transform_vector_to_y_up(x1, y1, x2, y2)
    return t1(x_data, y_data)


def get_ellipse_basic_data_by_three_point(x1, y1, x2, y2, x3, y3):
    return [[x1, y1], [x2, y2],
            get_distance_by_tow_point(x1, y1, x3, y3) + get_distance_by_tow_point(x2, y2, x3, y3)]


def get_ellipse_basic_data_by_isosceles_triangle(p1x, p1y, p2x, p2y, b):
    ox, oy = (p1x + p2x) / 2, (p1y + p2y) / 2
    l1 = get_distance_by_tow_point(p1x, p1y, p2x, p2y)
    a = l1 / 2
    if b < a:
        cc = a ** 2 - b ** 2
        c = sqrt(cc)
        c_a = c / a
        cx1 = ox + c_a * (p1x - ox)
        cy1 = oy + c_a * (p1y - oy)
        cx2 = 2 * ox - cx1
        cy2 = 2 * oy - cy1
        return [[cx1, cy1], [cx2, cy2], l1]
    else:
        cc = b ** 2 - a ** 2
        c = sqrt(cc)
        add = 2 * b
        if p1y == p2y:
            return [[ox, p1y + c], [ox, p1y - c], add]
        k = (p2x - p1x) / (p1y - p2y)
        x = c / sqrt(1 + k ** 2)
        cx1 = x + ox
        cy1 = k * x + oy
        cx2 = 2 * ox - cx1
        cy2 = 2 * oy - cy1
        return [[cx1, cy1], [cx2, cy2], add]



from ..core.king_decorater import geometry_generate
from .king_conic import KingConic
from ..alg.conic_alg import get_conic_mat_by_five_point, get_conic_coefficient
from ..alg.basic_alg import rref


def _five_point_conic(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5):
    mat = get_conic_mat_by_five_point(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5)
    mat = rref(mat)
    coefficient = get_conic_coefficient(mat)

    return coefficient


@geometry_generate(cls=KingConic)
def five_point_conic(depends=(), essential_data=None):
    x1, y1 = depends[0].basic_data
    x2, y2 = depends[1].basic_data
    x3, y3 = depends[2].basic_data
    x4, y4 = depends[3].basic_data
    x5, y5 = depends[4].basic_data

    return _five_point_conic(x1, y1, x2, y2, x3, y3, x4, y4, x5, y5)



from ..alg.circle_alg import get_circle_xy_data_by_center_radius
from ..alg.ellipse_alg import get_ellipse_xy_data_by_focus_add
from ..alg.hyperbola_alg import get_hyperbola_xy_data_by_focus_diff
from ..core.king_figure import KingFigure
from .king_curve import KingCurve
from ..alg.parabola_alg import get_parabola_xy_data_by_focus_point
from ..alg.conic_alg import CONIC_TYPE_CIRCLE, CONIC_TYPE_ELLIPSE, CONIC_TYPE_HYPERBOLA, CONIC_TYPE_PARABOLA


class KingCircle(KingCurve):
    conic_type_cls_var = CONIC_TYPE_CIRCLE

    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingCircle, self).__init__(depends=depends, essential_data=essential_data,
                                         create_type=create_type)

    def generate_basic_data(self, depends=(), essential_data=None):
        return [self.depends[0].basic_data, self.essential_data[0]]

    def generate_xy_data_from_basic_data(self):
        x1, y1 = self.basic_data[0]
        r = self.basic_data[1]
        return get_circle_xy_data_by_center_radius(x1, y1, r)


class KingEllipse(KingCurve):
    conic_type_cls_var = CONIC_TYPE_ELLIPSE

    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingEllipse, self).__init__(depends=depends, essential_data=essential_data,
                                          create_type=create_type)

    def generate_basic_data(self, depends=(), essential_data=None):
        return [self.depends[0].basic_data, self.depends[1].basic_data, self.essential_data[0]]

    def generate_xy_data_from_basic_data(self):
        x1, y1 = self.basic_data[0]
        x2, y2 = self.basic_data[1]
        b2 = self.basic_data[2]
        return get_ellipse_xy_data_by_focus_add(x1, y1, x2, y2, b2)


class KingHyperbola(KingCurve):
    conic_type_cls_var = CONIC_TYPE_HYPERBOLA

    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingHyperbola, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                            infinite=True)

    def generate_basic_data(self, depends=(), essential_data=None):
        return [depends[0].basic_data, depends[1].basic_data, essential_data[0]]

    def generate_xy_data_from_basic_data(self):
        x1 = self.basic_data[0][0]
        y1 = self.basic_data[0][1]
        x2 = self.basic_data[1][0]
        y2 = self.basic_data[1][1]
        b2 = self.basic_data[2]
        x_min, x_max = KingFigure().get_x_lim()
        y_min, y_max = KingFigure().get_y_lim()

        return get_hyperbola_xy_data_by_focus_diff(x1, y1, x2, y2, b2, x_min, x_max, y_min, y_max)


class KingParabola(KingCurve):
    conic_type_cls_var = CONIC_TYPE_PARABOLA

    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingParabola, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                           infinite=True)

    def generate_basic_data(self, depends=(), essential_data=None):
        return [depends[0].basic_data, depends[1].basic_data]

    def generate_xy_data_from_basic_data(self):
        xp, yp = self.basic_data[0]
        xv, yv = self.basic_data[1]
        x_min, x_max = KingFigure().get_x_lim()
        y_min, y_max = KingFigure().get_y_lim()

        return get_parabola_xy_data_by_focus_point(xp, yp, xv, yv, x_min, x_max, y_min, y_max)



from ..alg.plane_geometry_alg import get_distance_by_tow_point


def get_line_segment_length(line_segment):
    p1 = line_segment.basic_data[0]
    p2 = line_segment.basic_data[1]
    return get_distance_by_tow_point(p1[0], p1[1], p2[0], p2[1])


def attribute_line_segment_length_when_length_change(line_segment):
    line_segment.len = 0

    def inner(line):
        line.len = get_line_segment_length(line)
    line_segment.append_call_back(inner)


def print_line_segment_length_when_length_change(line_segment):
    def inner(line):
        print(get_line_segment_length(line))
    line_segment.append_call_back(inner)


def label_line_segment_length_when_length_change(line_segment):
    def inner(line):
        line.text.set_text(format(get_line_segment_length(line), ".2f"))
    line_segment.append_call_back(inner)



from ..core.king_geometry import KingGeometry
from matplotlib.pyplot import Artist
import numpy as np
from ..core.king_figure import KingFigure


class KingEquation(KingGeometry, Artist):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        self.contour = None
        super(KingGeometry, self).__init__()
        super(KingEquation, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                           infinite=True)
        self.mouse_select_priority = 1

    def assign_basic_data_to_artist(self):
        x = np.linspace(*KingFigure().get_x_lim(), 100)
        y = np.linspace(*KingFigure().get_y_lim(), 100)
        z = self.basic_data[0](*np.meshgrid(x, y))
        if self.contour is not None:
            for c in self.contour.collections:
                c.remove()
            del self.contour
        self.contour = KingFigure().ax.contour(x, y, z, 0)

    def contains(self, mouseevent):
        return False, {}



from .basic_alg import linspace
from .plane_geometry_alg import get_distance_by_tow_point, get_transform_vector_to_y_up


def get_parabola_focus_by_standard_equation(a11, a12, a22, a1, a2, a0):
    if a11 == 0:
        return [-a1 / 2, 0]
    return [0, -a2 / 2]


def get_parabola_equation_coefficient_by_simplest_focus(xp, yp):
    if yp == 0:
        return [0, 0, 1, -2 * xp, 0, 0]
    return [1, 0, 0, 0, -2 * yp, 0]


def get_parabola_xy_data_by_standard_equation(a11, a12, a22, a1, a2, a0, x_min, x_max, y_min, y_max):
    if a22 == 0:
        x = linspace(x_min, x_max, 300)
        y = -a11 * x ** 2 / (2 * a2)
        return [x, y]
    y = linspace(y_min, y_max, 300)
    x = -a22 * y ** 2 / (2 * a1)
    return [x, y]


def get_parabola_xy_data_by_simplest_focus(xp, yp, x_min, x_max, y_min, y_max):
    return get_parabola_xy_data_by_standard_equation(
        *get_parabola_equation_coefficient_by_simplest_focus(xp, yp), x_min, x_max, y_min, y_max)


def get_parabola_equation_coefficient_by_focus_line(xp, yp, x1, y1, x2, y2):
    x1_x2 = x1 - x2
    y1_y2 = y1 - y2
    x1_x2_2 = x1_x2 ** 2
    y1_y2_2 = y1_y2 ** 2
    x1_x2_y1_y2 = x1_x2 * y1_y2
    dd = x1_x2_2 + y1_y2_2

    a11 = x1_x2_2 / dd
    a12 = x1_x2_y1_y2 / dd
    b1 = x1 - x1_x2_2 / dd * x1 - x1_x2_y1_y2 / dd * y1
    a22 = y1_y2_2 / dd
    b2 = y1 - x1_x2_y1_y2 / dd * x1 - y1_y2_2 / dd * y1

    a11 -= 1
    a22 -= 1

    return [a11 ** 2 + a12 ** 2 - 1,
            a11 * a12 + a12 * a22,
            a12 ** 2 + a22 ** 2 - 1,
            a11 * b1 + a12 * b2 + xp,
            a12 * b1 + a22 * b2 + yp,
            b1 ** 2 + b2 ** 2 - xp ** 2 - yp ** 2]


def get_parabola_xy_data_by_focus_point(xp, yp, xv, yv, x_min, x_max, y_min, y_max):
    p = get_distance_by_tow_point(xv, yv, xp, yp)
    t1, t2 = get_transform_vector_to_y_up(xp, yp, xv, yv)

    def f(x):
        return x ** 2 / (2 * p)

    center_x = (x_min + x_max) / 2
    center_y = (y_min + y_max) / 2
    x_lim = x_max - x_min
    center_x_new, center_y_new = t2(center_x, center_y)
    x_lim_new = 0.71 * x_lim

    x_set = linspace(center_x_new - x_lim_new, center_x_new + x_lim_new, 200)
    y_set = f(x_set)

    x_set_new, y_set_new = t1(x_set, y_set)

    return [x_set_new, y_set_new]



from ..core.king_geometry import KingGeometry
from matplotlib.pyplot import Artist


class EmptyGeometry(KingGeometry, Artist):
    def __init__(self, depends=(), essential_data=None, create_type=2):
        super(KingGeometry, self).__init__()
        super(EmptyGeometry, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                            infinite=True)
        self.set_visible(False)
        self.set_animated(False)
        self.set_picker(False)

    def contains(self, mouseevent):
        return False, {}



import functools


def singleton(cls):
    _instance = {}

    def _singleton(*args, **kwargs):
        if cls not in _instance:
            _instance[cls] = cls(*args, **kwargs)
        return _instance[cls]

    return _singleton


def geometry_generate(cls=None, calc_essential_data_by_delta=None):
    def actual_decorator(func):
        @functools.wraps(func)
        def process(depends=(), essential_data=None):
            obj = cls(depends=depends, essential_data=essential_data, create_type=2)
            obj.calc_essential_data_by_delta = calc_essential_data_by_delta
            obj.generate_basic_data = func
            obj.init()
            return obj

        return process

    return actual_decorator



__all__ = ['king_conic', 'king_conic_generator', 'king_conic_graph', 'king_conic_graph_generator', 'king_curve',
           'king_equation', 'king_function', 'king_parametric_equation', 'king_plane_geometry_graph',
           'king_plane_geometry_graph_generator', 'king_regular_polygon', 'king_slider',
           'king_triangle_center_generator']



from ..geometry.king_plane_geometry_graph import KingPoint
from ..core.king_decorater import geometry_generate
from ..core.king_figure import KingFigure


@geometry_generate(cls=KingPoint)
def _point_on_curve(depends=(), essential_data=None):
    curve = depends[0]
    x_data = curve.get_xdata()
    y_data = curve.get_ydata()
    essential_data = essential_data % len(x_data)
    return [x_data[essential_data], y_data[essential_data]]


def get_manhattan_distance(x1, y1, x2, y2):
    return abs(x1 - x2) + abs(y1 - y2)


def get_min_manhattan_distance_index_by_point_xy_data(x, y, x_data, y_data, i_from, i_range):
    index = None
    dis = None
    length = len(x_data)
    i_min = i_from - i_range
    i_max = i_from + i_range
    j = i_min
    while j < i_max:
        j2 = j % length
        new_dis = get_manhattan_distance(x, y, x_data[j2], y_data[j2])
        if dis is None:
            dis = new_dis
            index = j2
        else:
            if new_dis < dis:
                dis = new_dis
                index = j2
        j += 1
    return index


def get_index_by_curve_mouse(i, curve, delta_x, delta_y, new_x, new_y):
    x_data = curve.get_xdata()
    y_data = curve.get_ydata()
    length = len(x_data)
    i_range = 10 + int(abs((delta_x + delta_y) * 100))
    if i_range > length:
        i_range = length
    return get_min_manhattan_distance_index_by_point_xy_data(new_x, new_y, x_data, y_data, i, i_range)


def get_init_index(x_data, y_data, x_min=None, x_max=None, y_min=None, y_max=None):
    x_min_fig, x_max_fig = KingFigure().get_x_lim()
    y_min_fig, y_max_fig = KingFigure().get_y_lim()
    if x_min is None:
        x_min = x_min_fig
    if x_max is None:
        x_max = x_max_fig
    if y_min is None:
        y_min = y_min_fig
    if y_max is None:
        y_max = y_max_fig
    for i in range(len(x_data)):
        if x_min < x_data[i] < x_max and y_min < y_data[i] < y_max:
            return i
    return 0


def point_on_curve(curve, x_min=None, x_max=None, y_min=None, y_max=None):
    x_data = curve.get_xdata()
    y_data = curve.get_ydata()
    mouse_loc = get_init_index(x_data, y_data, x_min, x_max, y_min, y_max)
    res = _point_on_curve(depends=[curve], essential_data=mouse_loc)
    res.mouse_translation = True

    def new_calc_essential_data_by_delta(delta, new_data):
        ind = get_index_by_curve_mouse(res.essential_data, curve, delta[0], delta[1], new_data[0], new_data[1])
        return ind

    res.calc_essential_data_by_delta = new_calc_essential_data_by_delta

    return res



from ..core.king_figure import KingFigure
from .king_curve import KingCurve
from ..alg.plane_geometry_alg import get_edge_two_point_by_line


class KingPoint(KingCurve):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingPoint, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type)
        self.mouse_select_priority = 0
        self.set_marker('o')
        self.set_markersize(5)
        self.set_markerfacecolor('blue')
        self.set_markeredgecolor('black')

    def calc_essential_data_by_delta(self, delta, new_data):
        return [self.basic_data[0] + delta[0], self.basic_data[1] + delta[1]]

    def generate_xy_data_from_basic_data(self):
        return self.basic_data


class KingLine(KingCurve):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingLine, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type,
                                       infinite=True)

    def generate_basic_data(self, depends=(), essential_data=None):
        x1 = depends[0].basic_data[0]
        y1 = depends[0].basic_data[1]
        x2 = depends[1].basic_data[0]
        y2 = depends[1].basic_data[1]

        return get_edge_two_point_by_line(x1, y1, x2, y2, *KingFigure().get_x_lim(), *KingFigure().get_y_lim())

    def generate_xy_data_from_basic_data(self):
        return [[self.basic_data[0][0], self.basic_data[1][0]], [self.basic_data[0][1], self.basic_data[1][1]]]


class KingLineSegment(KingCurve):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingLineSegment, self).__init__(depends=depends, essential_data=essential_data, create_type=create_type)

    def generate_basic_data(self, depends=(), essential_data=None):
        return [self.depends[0].basic_data, self.depends[1].basic_data]

    def generate_xy_data_from_basic_data(self):
        return [[self.basic_data[0][0], self.basic_data[1][0]],
                [self.basic_data[0][1], self.basic_data[1][1]]]



from .alg.basic_alg import *
from .alg.circle_alg import *
from .alg.conic_alg import *
from .alg.ellipse_alg import *
from .alg.hyperbola_alg import *
from .alg.parabola_alg import *
from .alg.plane_geometry_alg import *
from .alg.triangle_center_alg import *
from .core.king_animation import *
from .core.king_decorater import *
from .core.king_decorater_multi_geometry import *
from .core.king_figure import *
from .core.king_geometry import *
from .geometry.king_conic import *
from .geometry.king_conic_generator import *
from .geometry.king_conic_graph import *
from .geometry.king_conic_graph_generator import *
from .geometry.king_curve import *
from .geometry.king_empty_geometry import *
from .geometry.king_equation import *
from .geometry.king_function import *
from .geometry.king_parametric_equation import *
from .geometry.king_plane_geometry_graph import *
from .geometry.king_plane_geometry_graph_generator import *
from .geometry.king_point_on_curve import point_on_curve
from .geometry.king_regular_polygon import *
from .geometry.king_triangle_center_generator import outer as ot
from .geometry.king_triangle_center_generator import median as md
from .geometry.king_triangle_center_generator import incenter as it
from .geometry.king_triangle_center_generator import orthocenter as oc
from .geometry_attribute.affine import *
from .geometry_attribute.angle import *
from .geometry_attribute.distance import *

__all__ = ['alg', 'core', 'geometry', 'geometry_attribute', 'asymptote', 'circle', 'conic', 'ellipse', 'equation',
           'function', 'hyperbola', 'incenter', 'intersect', 'line', 'line_segment', 'median', 'middle', 'orthocenter',
           'outer', 'parabola', 'parametric_equation', 'point', 'regular_polygon', 'rotate', 'symmetric',
           'symmetry_center', 'symmetry_lines', 'tangent_line', 'vertical']
conic_class_list = [KingConic, KingCircle, KingEllipse, KingHyperbola, KingParabola]
figure = KingFigure()
p_nan = KingPoint(essential_data=[nan, nan])


def point(x=None, y=None, p_list=None, label=None, color=None, from_number=1, curve=None,
          x_min=None, x_max=None, y_min=None, y_max=None):
    if isinstance(curve, KingCurve):
        p = point_on_curve(curve, x_min=x_min, x_max=x_max, y_min=y_min, y_max=y_max)
    elif x is not None and y is not None:
        p = KingPoint(essential_data=[x, y])
    elif p_list is not None:
        ps = []
        for i in p_list:
            ps.append(KingPoint(essential_data=i))
        if label is not None and from_number is not None:
            for i in range(len(ps)):
                ps[i].text.set_text(label + str(from_number + i))
        if color is not None:
            for p in ps:
                p.set_markerfacecolor(color)
                p.set_markeredgecolor(color)
        return ps
    else:
        p = None
    if label is not None:
        p.text.set_text(label)
    if color is not None:
        p.set_markerfacecolor(color)
        p.set_markeredgecolor(color)
    return p


def line(p1=None, p2=None, k=None, b=None, A=None, B=None, C=None):
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint):
        return KingLine(depends=[p1, p2])
    if k is not None and b is not None:
        return line_kb(essential_data=[k, b])
    if A is not None and B is not None and C is not None:
        return line_general(essential_data=[A, B, C])
    if k is not None and isinstance(p1, KingPoint):
        return line_point_k(depends=[p1], essential_data=[k])
    if k is not None and isinstance(p2, KingPoint):
        return line_point_k(depends=[p2], essential_data=[k])


def line_segment(p1, p2):
    return KingLineSegment(depends=[p1, p2])


def middle(o1, o2):
    if isinstance(o1, KingPoint) and isinstance(o2, KingPoint):
        return point_middle(depends=[o1, o2])


def rotate(o1, o2, t):
    if isinstance(o1, KingPoint) and isinstance(o2, KingPoint):
        return point_rotate(depends=[o1, o2], essential_data=[t])


def vertical(p, l):
    if isinstance(p, KingPoint) and isinstance(l, KingLine):
        return point_vertical_point_line(depends=[p, l])
    if isinstance(p, KingLine) and isinstance(l, KingPoint):
        return point_vertical_point_line(depends=[l, p])


def symmetry_center(o):
    if isinstance(o, KingConic):
        return point_conic_symmetry_center(depends=[o])


def symmetric(p, l):
    if isinstance(p, KingPoint):
        if isinstance(l, KingLine):
            return point_symmetric_point_line(depends=[p, l])
        if isinstance(l, KingPoint):
            return point_symmetric_point(depends=[p, l])
    if isinstance(p, KingLine) and isinstance(l, KingPoint):
        return point_symmetric_point_line(depends=[l, p])


def intersect(o1, o2):
    if isinstance(o1, KingLine):
        if isinstance(o2, KingLine):
            return point_intersect_by_two_line(depends=[o1, o2])
        if is_object_in_class_list(o2, conic_class_list):
            return point_intersect_line_conic(depends=[o1, o2])
        if isinstance(o2, KingLineSegment):
            return point_intersect_by_line_and_segment(depends=[o1, o2])
    if isinstance(o2, KingLine):
        if is_object_in_class_list(o1, conic_class_list):
            return point_intersect_line_conic(depends=[o2, o1])
        if isinstance(o1, KingLineSegment):
            return point_intersect_by_line_and_segment(depends=[o2, o1])
    if isinstance(o1, KingLineSegment):
        if isinstance(o2, KingLineSegment):
            return point_intersect_by_segment_and_segment(depends=[o1, o2])
    if isinstance(o1, KingCircle) and isinstance(o2, KingCircle):
        return point_intersect_by_circle_and_circle(depends=[o2, o1])


def asymptote(o):
    if isinstance(o, KingConic) or isinstance(o, KingHyperbola):
        return conic_asymptote(depends=[o])


def symmetry_lines(o):
    if isinstance(o, KingConic):
        return conic_symmetry_lines(depends=[o])


def tangent_line(p, o):
    if isinstance(o, KingConic) or is_object_in_class_list(o, conic_class_list):
        return tangent_line_by_conic(depends=[p, o])


def conic(coefficient=(), p_list=()):
    if len(coefficient) == 6:
        return KingConic(essential_data=coefficient)
    if len(p_list) == 5:
        return five_point_conic(depends=p_list)


def circle(o=None, r=None, p=None, p1=None, p2=None, p3=None):
    if isinstance(o, KingPoint):
        if r is not None:
            return KingCircle(depends=[o], essential_data=[r])
        if isinstance(p, KingPoint):
            return circle_center_edge_point(depends=[o, p])
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint) and isinstance(p3, KingPoint):
        return circle_three_point(depends=[p1, p2, p3])


def ellipse(e1=None, e2=None, p=None, add=None, v1=None, v2=None, b=None):
    if isinstance(e1, KingPoint) and isinstance(e2, KingPoint):
        if isinstance(p, KingPoint):
            return ellipse_three_point(depends=[e1, e2, p])
        if add is not None:
            return KingEllipse(depends=[e1, e2], essential_data=[add])
    if isinstance(v1, KingPoint) and isinstance(v2, KingPoint) and b is not None:
        return ellipse_isosceles_triangle(depends=[v1, v2], essential_data=[b])


def hyperbola(e1=None, e2=None, p=None, diff=None):
    if isinstance(e1, KingPoint) and isinstance(e2, KingPoint):
        if isinstance(p, KingPoint):
            return hyperbola_three_point(depends=[e1, e2, p])
        if diff is not None:
            return KingHyperbola(depends=[e1, e2], essential_data=[diff])


def parabola(o1, o2):
    if isinstance(o1, KingPoint):
        if isinstance(o2, KingPoint):
            return KingParabola(depends=[o1, o2])
        if isinstance(o2, KingLine):
            return parabola_point_line(depends=[o1, o2])
    if isinstance(o1, KingLine):
        if isinstance(o2, KingPoint):
            return parabola_point_line(depends=[o2, o1])


def orthocenter(p1, p2, p3):
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint) and isinstance(p3, KingPoint):
        return oc(depends=[p1, p2, p3])


def median(p1, p2, p3):
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint) and isinstance(p3, KingPoint):
        return md(depends=[p1, p2, p3])


def incenter(p1, p2, p3):
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint) and isinstance(p3, KingPoint):
        return it(depends=[p1, p2, p3])


def outer(p1, p2, p3):
    if isinstance(p1, KingPoint) and isinstance(p2, KingPoint) and isinstance(p3, KingPoint):
        return ot(depends=[p1, p2, p3])


def regular_polygon(o, p, n):
    if isinstance(o, KingPoint) and isinstance(p, KingPoint):
        return KingRegularPolygon(depends=[o, p], essential_data=[n])


def function(f, x_min=None, x_max=None, count=None):
    return KingFunction(essential_data={'function': f, 'x_min': x_min, 'x_max': x_max, 'count': count})


def equation(z):
    return KingEquation(essential_data=[z])


def parametric_equation(x, y, t_min, t_max, count=None):
    return KingParametricEquation(essential_data={'x': x, 'y': y, 't_min': t_min, 't_max': t_max, 'count': count})



import functools
from .king_decorater import geometry_generate
from ..geometry.king_empty_geometry import EmptyGeometry


def geometries_generate(classes=(), numbers=()):
    def actual_decorator(func):
        @functools.wraps(func)
        def process(depends=(), essential_data=None):
            @geometry_generate(cls=EmptyGeometry)
            def empty_geo_gen(depends=(), essential_data=None):
                return func(depends=depends, essential_data=essential_data)

            empty_geo = empty_geo_gen(depends=depends, essential_data=essential_data)
            geo_list = []
            s = 0
            for i in range(len(classes)):
                for j in range(numbers[i]):
                    @geometry_generate(cls=classes[i])
                    def every_geo_gen(depends=(), essential_data=None):
                        return depends[0].basic_data[essential_data[0]]

                    every_obj = every_geo_gen(depends=[empty_geo], essential_data=[s])
                    geo_list.append(every_obj)
                    s += 1
            return geo_list

        return process

    return actual_decorator



from ..core.king_geometry import KingGeometry
from matplotlib.pyplot import Artist
from ..alg.basic_alg import pi_2
from ..geometry.king_plane_geometry_graph_generator import point_rotate
from ..geometry.king_plane_geometry_graph import KingLineSegment


class KingRegularPolygon(KingGeometry, Artist):
    def __init__(self, depends=(), essential_data=None, create_type=1):
        super(KingGeometry, self).__init__()
        super(KingRegularPolygon, self).__init__(depends=depends, essential_data=essential_data,
                                                 create_type=create_type)
        self.mouse_select_priority = 1
        if create_type == 1:
            self.mouse_translation = True
        t = pi_2 / essential_data[0]
        i = 0
        v = [depends[1]]
        e = []
        self.vertex = v
        self.edge = e
        n = essential_data[0] - 1
        while i < n:
            v.append(point_rotate(depends=[depends[0], v[i]], essential_data=[t]))
            e.append(KingLineSegment(depends=[v[i], v[i + 1]]))
            i += 1
        e.append(KingLineSegment(depends=[v[0], v[i]]))

    def contains(self, mouseevent):
        return False, {}



from ..alg.plane_geometry_alg import get_angle_by_three_point_corr


def get_angle_by_three_point(p1, p2, p3):
    return get_angle_by_three_point_corr(*p1.basic_data, *p2.basic_data, *p3.basic_data)


def print_angle_by_three_point_when_change(p1, p2, p3):
    p1.append_call_back(lambda point_1: print(get_angle_by_three_point(point_1, p2, p3)))
    p2.append_call_back(lambda point_2: print(get_angle_by_three_point(p1, point_2, p3)))
    p3.append_call_back(lambda point_3: print(get_angle_by_three_point(p1, p2, point_3)))


def label_angle_by_three_point_when_change(p1, p2, p3):

    def inner1(point):
        p2.text.set_text(format(get_angle_by_three_point(point, p2, p3), ".2f"))

    def inner2(point):
        p2.text.set_text(format(get_angle_by_three_point(p1, point, p3), ".2f"))

    def inner3(point):
        p2.text.set_text(format(get_angle_by_three_point(p1, p2, point), ".2f"))

    p1.append_call_back(inner1)
    p2.append_call_back(inner2)
    p3.append_call_back(inner3)



import abc
from .king_figure import KingFigure
from matplotlib.pyplot import Artist


class KingGeometry(metaclass=abc.ABCMeta):
    def __init__(self, depends=(), essential_data=None, create_type=1, infinite=False):
        if isinstance(self, Artist):
            self.set_animated(True)
            self.set_picker(True)
            KingFigure().add_artist(self)

        # Call backs when invoke update_self.
        self.call_backs = []

        # Update search flag.
        self.searched = False

        # Depends KingGeometry object list.
        self.depends = depends

        # Essential data supplement for depends.
        self._essential_data = essential_data

        # Basic data for assign Artist's init method.
        self._basic_data = None

        # KingGeometry who depends on this.
        self.feeds = []

        # A KingGeometry object can be instanced by tow methods.
        # 1: Free (can by moved by mouse)
        # 2: NoFree (can not move directly by mouse)
        self.create_type = create_type

        # Allow mouse translation. default False
        self.mouse_translation = False

        # Mouse select priority. The smallest first.
        self.mouse_select_priority = 10

        if create_type == 1:
            self.init()
        if infinite:
            KingFigure().inf_geometry.append(self)

    def append_call_back(self, f):
        self.call_backs.append(f)
        f(self)

    @property
    def basic_data(self):
        return self._basic_data

    @basic_data.setter
    def basic_data(self, basic_data):
        self._basic_data = basic_data

    @property
    def essential_data(self):
        return self._essential_data

    def set_depends_feeds(self):
        for node in self.depends:
            node.__add_feeds(self)

    def __add_feeds(self, node):
        self.feeds.append(node)

    def init(self):
        self.set_depends_feeds()
        self._basic_data = self.generate_basic_data(self.depends, self._essential_data)
        self.assign_basic_data_to_artist()
        if not KingFigure().init_delay_update:
            KingFigure().update()

    def generate_basic_data(self, depends=(), essential_data=None):
        """
            Default method to generate basic data.
        you can not use the parameter. you can only use self attribute.
        you might override this.
        :return:
        """
        return self._essential_data

    def assign_basic_data_to_artist(self):
        """
            Assign basic data to Artist's derived classes.
        :return:
        """
        pass

    def update_self(self, new_essential_data=None):
        """
            Update self.
        :return:
        """
        if new_essential_data is not None:
            self._essential_data = new_essential_data
        self._basic_data = self.generate_basic_data(self.depends, self._essential_data)
        self.assign_basic_data_to_artist()
        for func in self.call_backs:
            func(self)

    def update_affected(self, new_essential_data=None, repaint=True):
        """
            Update affected.
        :return: All affected artists.
        """
        KingFigure().clear_searched()

        self.update_self(new_essential_data)
        self.searched = True
        result = [self]
        queue = []
        for f in self.feeds:
            if not f.searched:
                queue.append(f)
                result.append(f)
                f.searched = True

        while len(queue) > 0:
            head = queue.pop(0)
            for f in head.feeds:
                if not f.searched:
                    queue.append(f)
                    result.append(f)
                    f.searched = True
            head.update_self()
        if repaint:
            KingFigure().update()
        return result

    def calc_essential_data_by_delta(self, delta, new_data):
        """
            Calculate essential data by delta.
        :param delta:
        :param new_data:
        :return: essential data
        """
        return self._essential_data

    def translation(self, delta, new_data):
        """
            Translation method. It can translation by mouse if and only if the depends geometry has no depends.
            Parameters
            ----------
            delta : (float, float)
                    delta x and delta y.
        :return:
        """
        ready2update = []
        ready2update.extend(self.depends)
        ready2update.append(self)
        KingFigure().clear_searched()
        for a in ready2update:
            if a.calc_essential_data_by_delta is None:
                esd = None
            else:
                esd = a.calc_essential_data_by_delta(delta, new_data)
            a.update_self(esd)
            a.searched = True
        for b in ready2update:
            queue = []
            for f in b.feeds:
                if not f.searched:
                    queue.append(f)
                    f.searched = True
            while len(queue) > 0:
                head = queue.pop(0)
                for f in head.feeds:
                    if not f.searched:
                        queue.append(f)
                        f.searched = True
                head.update_self()



from matplotlib.widgets import Slider, Button
import matplotlib.pyplot as plt

# TODO: slider

plt.subplots_adjust(left=0.25, bottom=0.25)

axamp = plt.axes([0.1, 0.25, 0.0225, 0.63], facecolor='lightgoldenrodyellow')
amp_slider = Slider(
    ax=axamp,
    label="Amplitude",
    valmin=0,
    valmax=10,
    valinit=5,
    orientation="vertical"
)


def update(val):
    print(val)


amp_slider.on_changed(update)

resetax = plt.axes([0.8, 0.025, 0.1, 0.04])
button = Button(resetax, 'Reset', color='lightgoldenrodyellow', hovercolor='0.975')


def reset(event):
    amp_slider.reset()


button.on_clicked(reset)

plt.show()



from src import king

count = 14


@king.geometries_generate(classes=[king.KingPoint], numbers=[count])
def point_intersect_line_conic(depends=(), essential_data=None):
    c = depends[0].basic_data[0]
    a_list = [(0, 0)]

    for i in range(count):
        a_list.append((i + 1, a_list[i][1] ** 2 / 2 + c))
        print(i)

    return a_list


cp = king.point(1, -3)
point_intersect_line_conic(depends=[cp])
